"""
###########################
MIPPY: Modular Image Processing in Python
###########################
mippy.application
Author: Robert Flintham

This file contains MIPPY's main "Application" class. This defines the main
DICOM browser window, and what each of the menu options/buttons etc
all do.
"""

# Import system/python modules
#~ print "Importing system modules"
from pkg_resources import resource_filename, parse_version
import os
import tkinter.messagebox
import tkinter.filedialog
from tkinter import *
from tkinter.ttk import *
#~ import ScrolledText
from datetime import datetime
import sys
import numpy as np
import time
import webbrowser
import shutil
import importlib
import getpass
import pickle as pickle
import itertools
from functools import partial
import stat
import pydicom
import gc
import requests
import csv
#~ from multiprocessing import freeze_support
#~ print "Imports finished!"

#~ print "Importing MIPPY code"
from . import viewing as mview
from . import mdicom
from .mdicom.reading import collect_dicomdir_info
from .mdicom.reading import get_dataset
from .mdicom.reading import compare_dicom
from .mdicom.reading import load_images_from_uids
from .mdicom.mrenhanced import get_frame_ds
from .mdicom.io import save_temp_ds
from . import fileio
from .threading import multithread
import imp
#~ print "Done!"
from zipfile import ZipFile

# WEB LINKS
MIPPYDOCS = r'http://mippy.readthedocs.io'


class MIPPYMain(Frame):

    ## Define any global attributes/variables you'll need here.  Use as few of these
    ## as possible as they will exist in RAM for as long as the program is running
    ## and are a waste of memory if they're not needed.  They also create massive
    ## confusion if you want to use similar/the same variable names in other modules






    def __init__(self, master):
        """
        This is the function which is called automatically when you create an object of
        this class, to initiate ("init", get it?) the object.

        This is where the GUI is first created and populated with buttons, image canvasses,
        tabs, whatever.  This function can also call other functions should you wish to split
        the code down further.
        """

        # IMPORT STATEMENTS - THESE ARE NOT CALLED UNTIL THE APPLICATION IS STARTED



        self.root_dir = os.getcwd()
        self.dicomdir = None

        # Initialises the GUI as a "Frame" object and gives it the name "master"
        Frame.__init__(self, master)
        self.master = master
        self.master.root_dir = os.getcwd()
        self.master.title("MIPPY: Modular Image Processing in Python")
        #~ self.master.minsize(650,400)
        #~ self.root_path = os.getcwd()



        if "nt" == os.name:
            impath = resource_filename('mippy','resources/brain_orange.ico')
        else:
            impath = '@'+resource_filename('mippy','resources/brain_bw.xbm')
        self.master.wm_iconbitmap(impath)

        # Catches any calls to close the window (e.g. clicking the X button in Windows) and pops
        # up an "Are you sure?" dialog
        self.master.protocol("WM_DELETE_WINDOW", self.asktoexit)








        '''
        This section is to determine the host OS, and set up the appropriate temp
        directory for images.
        Windows = C:\Temp
        Mac = /tmp
        Linux = /tmp
        '''



        self.user = getpass.getuser()

        # Collect proxy information for requests
        self.proxy_list = []
        if os.path.exists('proxies.csv'):
            with open('proxies.csv','r') as proxy_csv:
                reader = csv.DictReader(proxy_csv)
                for row in reader:
                    self.proxy_list.append(dict(row))
            print(self.proxy_list)


        # Check if frozen:
        import sys
        if getattr(sys, 'frozen', False):
            self.frozen = True
        else:
            self.frozen = False

        # Use parallel processing?
        if not self.frozen:
            self.multiprocess = True
        else:
            self.multiprocess = False

        print("Multiprocessing is {}".format(self.multiprocess))


        import pkg_resources
        if not self.frozen:
            if sys.argv[0]=='mippydev.py':
                self.mippy_version = "DEVELOPMENT_VERSION"
            else:
                self.mippy_version = pkg_resources.get_distribution("mippy").version
        else:
            self.mippy_version = "FROZEN - UNKNOWN"

        # Set temp directory
        if 'darwin' in sys.platform or 'linux' in sys.platform:
            self.tempdir = '/tmp/MIPPY_temp_'+self.user
        elif 'win' in sys.platform:
            fallback_temp = r'C:\TEMP'
            sys_temp  = os.getenv("TEMP",fallback_temp)

            # If sys_temp not accessible, default to fallback_temp
            if not os.access(sys_temp, os.W_OK):
                sys_temp = fallback_temp
                print("System reported temp directory not available - using {}".format(sys_temp))

            self.tempdir = os.path.join(sys_temp,"MIPPY_temp_"+self.user)
        else:
            tkinter.messagebox.showerror('ERROR', 'Unsupported operating system, please contact the developers.')
            sys.exit()
        if not os.path.exists(self.tempdir):
            os.makedirs(self.tempdir)

        # Set persistent user directory
        self.fallback_userdir = None
        self.userdir = None
        if 'darwin' in sys.platform or 'linux' in sys.platform:
            #~ self.userdir = os.path.join('/home',self.user,'.mippy')
            self.userdir = os.path.expanduser('~/.mippy')
        elif 'win' in sys.platform:
            fallback_sys_userdir = os.path.join(r'C:\Users',self.user)
            sys_userdir = os.getenv('APPDATA',fallback_sys_userdir)

            # If sys_userdir not available, default to fallback_userdir
            if not os.access(sys_userdir, os.W_OK):
                sys_userdir = fallback_sys_userdir
                print("System reported user directory not available - using {}".format(sys_userdir))

            self.userdir = os.path.join(sys_userdir,'.mippy')
            self.fallback_userdir = os.path.join(fallback_sys_userdir,'.mippy')
        else:
            tkinter.messagebox.showerror('ERROR', 'Unsupported operating system, please contact the developers.')
            sys.exit()

        if not os.path.exists(self.userdir):
            os.makedirs(self.userdir)

        # If newer files in fallback directory than system reported directory, update files
        if ('win' in sys.platform
            and not 'darwin' in sys.platform
            and os.path.exists(self.fallback_userdir)
            and not self.fallback_userdir==self.userdir):
            files_updated = fileio.copyToDir(self.fallback_userdir,self.userdir,ignore='*.json')
            if len(files_updated)>0:
                print("The following fles have been updated in your home directory:")
                for f in files_updated:
                        print("- {}".format(f))

        # Set default module directory
        if os.path.exists(os.path.join(self.userdir,'defaultmodulepath.txt')):
            with open(os.path.join(self.userdir,'defaultmodulepath.txt'),'r') as modpathfile:
                requested_moduledir = modpathfile.readlines()[0].lstrip().rstrip()
                if os.path.exists(requested_moduledir):
                    self.moduledir = requested_moduledir
                    print('USING MODULE DIRECTORY: {}'.format(self.moduledir))
        elif os.path.exists(os.path.join(self.root_dir, 'modules')):
            self.moduledir = os.path.join(self.root_dir, 'modules')
            if not self.moduledir in sys.path:
                sys.path.append(self.moduledir)
        else:
            self.moduledir = None

        # Create a list to hold module eggs
        self.module_eggs = []

        # Create variable for export directory
        self.exportdir = None

        # Check status of DCMDJPEG for mac or unix, and set
        # executable if necessary

        # Create "working copy" of the correct DCMDJPEG in the temp folder and
        # set executable flag as necessary

        if 'darwin' in sys.platform:
            dcmdjpegpath = resource_filename('mippy','resources/dcmdjpeg_mac')
        elif 'linux' in sys.platform:
            dcmdjpegpath = resource_filename('mippy','resources/dcmdjpeg_linux')
        elif 'win' in sys.platform:
            dcmdjpegpath = resource_filename('mippy','resources/dcmdjpeg_win.exe')

        dcmdjpeg_copy = os.path.join(self.tempdir,os.path.split(dcmdjpegpath)[1])

        shutil.copy(dcmdjpegpath,dcmdjpeg_copy)

        #~ print os.stat(dcmdjpeg_copy)

        if 'darwin' in sys.platform or 'linux' in sys.platform:
            st = os.stat(dcmdjpeg_copy)
            os.chmod(dcmdjpeg_copy,st.st_mode | stat.S_IEXEC)

        ## Change ttk theme on linux

        if 'darwin' in sys.platform:
            pass
        elif 'linux' in sys.platform:
            s = Style()
            s.theme_use('default')
        elif 'win' in sys.platform:
            dcmdjpegpath = resource_filename('mippy','resources/dcmdjpeg_win.exe')

        # Create menu bar for the top of the window
        self.menubar = Menu(master)
        # Create and populate "File" menu
        self.filemenu = Menu(self.menubar, tearoff=0)
        self.filemenu.add_command(label="Load new image directory", command=lambda:self.load_image_directory())
        self.filemenu.add_command(label="Open user home directory", command=lambda:self.open_user_directory())
        self.filemenu.add_command(label="Exit program",command=lambda:self.exit_program())
        # Create and populate "Modules" menu
        self.modulemenu = Menu(self.menubar,tearoff=0)
        self.modulemenu.add_command(label="Load new module directory",command=lambda:self.select_modules_directory())
        self.modulemenu.add_command(label="Refresh module list", command=lambda:self.scan_modules_directory())
        # Create and populate "Image" menu
        self.imagemenu = Menu(self.menubar,tearoff=0)
        self.imagemenu.add_command(label="View DICOM header", command=lambda:self.view_header())
        self.imagemenu.add_command(label="Compare DICOM headers", command=lambda:self.compare_headers())
        self.imagemenu.add_command(label="Export DICOM files", command=lambda:self.export_dicom())
        # Create and populate "Options" menu
        self.optionsmenu = Menu(self.menubar,tearoff=0)
        self.optionsmenu.add_command(label="Enable multiprocessing", command=lambda:self.enable_multiprocessing())
        self.optionsmenu.add_command(label="Disable multiprocessing", command=lambda:self.disable_multiprocessing())
        # Create and populate "Help" menu
        self.helpmenu = Menu(self.menubar, tearoff=0)
        self.helpmenu.add_command(label="Open MIPPY documentation",command=lambda:self.load_docs())
        self.helpmenu.add_command(label="About MIPPY",command=lambda:self.display_version_info())
        #~ self.helpmenu.add_command(label="View changelog",command=lambda:self.display_changelog())
        self.helpmenu.add_command(label="Report issue",command=lambda:self.report_issue())
        #~ self.helpmenu.add_command(label="View current log file",command=lambda:self.show_log())
        # Add menus to menubar and display menubar in window
        self.menubar.add_cascade(label="File",menu=self.filemenu)
        self.menubar.add_cascade(label="Modules",menu=self.modulemenu)
        self.menubar.add_cascade(label="Image",menu=self.imagemenu)
        self.menubar.add_cascade(label="Options",menu=self.optionsmenu)
        self.menubar.add_cascade(label="Help",menu=self.helpmenu)
        self.master.config(menu=self.menubar)

        # Create frames to hold DICOM directory tree and module list
        self.dirframe = Frame(self.master)
        self.moduleframe = Frame(self.master)
        self.dirframe = Frame(self.master)
        self.moduleframe = Frame(self.master)

        # Create DICOM treeview
        self.dirframe.dicomtree = Treeview(self.dirframe)

        # Set names and widths of columns in treeviews
        self.dirframe.dicomtree['columns']=('date','name','desc')
        self.dirframe.dicomtree.heading('date',text='Study Date')
        self.dirframe.dicomtree.heading('name',text='Patient Name')
        self.dirframe.dicomtree.heading('desc',text='Description')
        self.dirframe.dicomtree.column('#0',width=100,stretch=False)
        self.dirframe.dicomtree.column('date',width=100,stretch=False)
        self.dirframe.dicomtree.column('name',width=200)
        self.dirframe.dicomtree.column('desc',width=500)

        # Create scrollbars
        self.dirframe.scrollbarx = Scrollbar(self.dirframe,orient='horizontal')
        self.dirframe.scrollbarx.config(command=self.dirframe.dicomtree.xview)
        self.dirframe.scrollbary = Scrollbar(self.dirframe)
        self.dirframe.scrollbary.config(command=self.dirframe.dicomtree.yview)
        self.dirframe.dicomtree.configure(yscroll=self.dirframe.scrollbary.set, xscroll=self.dirframe.scrollbarx.set)

        # Use "grid" to postion treeview and scrollbars in DICOM frame and assign weights to columns and rows
        self.dirframe.dicomtree.grid(row=0,column=0,sticky='nsew')
        self.dirframe.scrollbarx.grid(row=1,column=0,sticky='nsew')
        self.dirframe.scrollbary.grid(row=0,column=1,sticky='nsew')

        # Set "weights" (relatve amount of stretchability when resizing window) for each row and column
        self.dirframe.rowconfigure(0,weight=1)
        self.dirframe.columnconfigure(0,weight=1)
        self.dirframe.rowconfigure(1,weight=0)
        self.dirframe.columnconfigure(1,weight=0)

        # Bind "change selection" event to method to update the image display
        self.dirframe.dicomtree.bind('<<TreeviewSelect>>',self.dir_window_selection)

        # Remove focus from dicomtree widget when mouse not hovering
        #~ self.master.dirframe.dicomtree.bind('<Leave>',self.dicomtree_nofocus)
        #~ self.master.dirframe.dicomtree.bind('<Enter>',self.dicomtree_focus)


        # Create module treeview
        self.moduleframe.moduletree = Treeview(self.moduleframe)

        # Set names and widths of columns in treeview
        self.moduleframe.moduletree['columns']=('description','author','version')
        self.moduleframe.moduletree.heading('#0',text='Module Name')
        self.moduleframe.moduletree.heading('description',text='Description')
        self.moduleframe.moduletree.heading('author',text='Author')
        self.moduleframe.moduletree.heading('version',text='Version')

        # Create scrollbars
        self.moduleframe.scrollbarx = Scrollbar(self.moduleframe,orient='horizontal')
        self.moduleframe.scrollbarx.config(command=self.moduleframe.moduletree.xview)
        self.moduleframe.scrollbary = Scrollbar(self.moduleframe)
        self.moduleframe.scrollbary.config(command=self.moduleframe.moduletree.yview)
        self.moduleframe.moduletree.configure(yscroll=self.moduleframe.scrollbary.set, xscroll=self.moduleframe.scrollbarx.set)

        # Use "grid" to postion treeview and scrollbars in DICOM frame and assign weights to columns and rows
        self.moduleframe.moduletree.grid(row=0,column=0,sticky='nsew')
        self.moduleframe.scrollbarx.grid(row=1,column=0,sticky='nsew')
        self.moduleframe.scrollbary.grid(row=0,column=1,sticky='nsew')

        # Set "weights" (relatve amount of stretchability when resizing window) for each row and column
        self.moduleframe.rowconfigure(0,weight=1)
        self.moduleframe.columnconfigure(0,weight=1)
        self.moduleframe.rowconfigure(1,weight=0)
        self.moduleframe.columnconfigure(1,weight=0)

        # Remove focus from moduletree widget when mouse not hovering
        #~ self.master.moduleframe.moduletree.bind('<Leave>',self.moduletree_nofocus)
        #~ self.master.moduleframe.moduletree.bind('<Enter>',self.moduletree_focus)

        # Load modules to list
        self.scan_modules_directory()
        # TEMPORARILY DISABLED

        # Bind "module select" event to required action
        self.moduleframe.moduletree.bind('<<TreeviewSelect>>',self.module_window_click)

        # Just adding a random line to the tree for testing
        #self.master.moduleframe.moduletree.insert('','end',"test row",text="Blah blah",values=("Option 1","Option 2"))

        # Create canvas object to draw images in
        self.imcanvas = mview.MIPPYCanvas(self.master,bd=0,width=256, height=256,drawing_enabled=False)


        # Add a scrollbar to MIPPYCanvas to enable slice scrolling
        self.imcanvas.img_scrollbar = Scrollbar(self.master,orient='horizontal')
        self.imcanvas.configure_scrollbar()

        # Create buttons:
        # "Load module"
        self.loadmodulebutton = Button(self.master,text="Load module",command=lambda:self.load_selected_module())

        # Add progressbar
        self.master.progressbar = Progressbar(self.master, mode='determinate')

        # Use "grid" to position objects within "master"
        self.dirframe.grid(row=0,column=0,columnspan=2,rowspan=1,sticky='nsew')
        self.imcanvas.grid(row=1,column=0,sticky='nsew')
        self.moduleframe.grid(row=1,column=1,sticky='nsew')
        self.loadmodulebutton.grid(row=2,column=1,sticky='nsew')
        #~ self.scrollbutton.grid(row=2,column=0,sticky='nsew')
        self.imcanvas.img_scrollbar.grid(row=2,column=0,sticky='ew')
        self.master.progressbar.grid(row=3,column=0,rowspan=1,columnspan=2,sticky='nsew')

        # Set row and column weights to handle resizing
        self.master.rowconfigure(0,weight=1)
        self.master.rowconfigure(1,weight=0)
        self.master.rowconfigure(2,weight=0)
        self.master.rowconfigure(3,weight=0)
        self.master.columnconfigure(0,weight=0)
        self.master.columnconfigure(1,weight=1)

        self.focus()

        # Here are some variables that may be useful
        self.open_ds = None
        self.open_file = None

    def slice_scroll_button_click(self,event):
        self.click_x = event.x
        self.click_y = event.y
        #~ print "CLICK"
        return

    def open_user_directory(self):
        webbrowser.open(self.userdir)
        return




    def asktoexit(self):
#        if tkMessageBox.askokcancel("Quit?", "Are you sure you want to exit?"):
#            self.master.destroy()
            #~ sys.exit()
        mbox = tkinter.messagebox.Message(
            title='Delete temporary files?',
            message='Would you like to delete all MIPPY temp files?',
            icon=tkinter.messagebox.QUESTION,
            type=tkinter.messagebox.YESNOCANCEL,
            master = self)
        reply = mbox.show()
        if reply=='yes':
            self.clear_temp_dir()
            self.master.destroy()
        elif reply=='no':
            self.master.destroy()
        else:
            return
        return



    def dir_window_selection(self,event):
        # THIS NEEDS IF len==1 to decide how to draw preview images
        selection = self.dirframe.dicomtree.selection()
        self.active_uids = []
        for item in selection:
            parent_item = self.dirframe.dicomtree.parent(item)
            if parent_item=='':
                # Whole study, not sure what to do...
                self.imcanvas.reset()
                #~ print "Whole study selected, no action determined yet."
            elif self.dirframe.dicomtree.parent(parent_item)=='':
                # Whole series, add children to list
                for image_uid in self.dirframe.dicomtree.get_children(item):
                    self.active_uids.append(image_uid)
                if len(selection)==1:
                    if not item==self.active_series:
                        self.load_preview_images(self.dirframe.dicomtree.get_children(item))
                        self.active_series = item
                    self.imcanvas.show_image(1)
            else:
                # Single slice
                self.active_uids.append(item)
                if len(selection)==1:
                    parent = self.dirframe.dicomtree.parent(item)
                    if not parent==self.active_series:
                        self.load_preview_images(self.dirframe.dicomtree.get_children(parent))
                        self.active_series = parent
                    self.imcanvas.show_image(self.dirframe.dicomtree.index(item)+1)

    def progress(self,percentage):
        self.master.progressbar['value']=percentage
        self.master.progressbar.update()

    def load_preview_images(self, uid_array):
        """
        Requires an array of unique instance UID's to search for in self.tag_list
        """
        #~ self.reset_small_canvas()
        n = 0
        preview_images = []
        for tag in self.sorted_list:
            if tag['instanceuid'] in uid_array:
                self.progress(10.*n/len(uid_array))
                preview_images.append(tag['px_array'])
                n+=1
        self.imcanvas.load_images(preview_images)
        return



    def module_window_click(self,event):
        # print("You clicked on the module window.")
        pass


    def load_image_directory(self):
        print("Load image directory")
        prev_dir = self.dicomdir
        try:
            # Unpickle previous directory if available
            with open(os.path.join(self.userdir,'prevdir.cfg'),'rb') as cfg_file:
                prevdir = pickle.load(cfg_file)
            print("PREV DIRECTORY: {}".format(prevdir))
        except:
            prevdir = None
        self.dicomdir = tkinter.filedialog.askdirectory(parent=self,initialdir=prevdir,title="Select image directory")
        if not self.dicomdir:
            self.dicomdir = prev_dir
            return
        with open(os.path.join(self.userdir,'prevdir.cfg'),'wb') as cfg_file:
            pickle.dump(self.dicomdir,cfg_file)
        self.ask_recursive = tkinter.messagebox.askyesno("Search recursively?","Do you want to include all subdirectories?")

        #~ self.path_list = []
        self.active_series = None

        # Check for appropriate mippydb object in the chosen directory
        if self.ask_recursive:
            self.mippydbpath = os.path.join(self.dicomdir,"mippydb_r")
        else:
            self.mippydbpath = os.path.join(self.dicomdir,"mippydb")
        if os.path.exists(self.mippydbpath):
            ask_use_mippydb = tkinter.messagebox.askyesno("Load existing MIPPYDB?","MIPPYDB file found. Load the existing DICOM tree?")
            if ask_use_mippydb:
                with open(self.mippydbpath,'rb') as fileobj:
                    self.sorted_list = pickle.load(fileobj)
                self.tag_list = self.sorted_list    # Shouldn't need this really...
                self.build_dicom_tree()
                return

        self.path_list = fileio.list_all_files(self.dicomdir,recursive=self.ask_recursive)

        self.filter_dicom_files()
        self.build_dicom_tree()

        return

    def filter_dicom_files(self):
        self.tag_list = []
        errors = []
        if self.multiprocess and not (('win' in sys.platform and not 'darwin' in sys.platform)
                                and len(self.path_list)<20):
            f = partial(collect_dicomdir_info,tempdir=self.tempdir)
            self.tag_list = multithread(f,self.path_list,progressbar=self.progress)

            self.tag_list = [item for sublist in self.tag_list for item in sublist]
            self.tag_list = [value for value in self.tag_list if value != []]
            errors = [value for value in self.tag_list if value['state']=='ERROR']
            # self.tag_list = [value for value in self.tag_list if not value in errors]
            self.tag_list = [value for value in self.tag_list if value != None and not value['state']=='ERROR']
        else:
            for p in self.path_list:
                self.progress(100*(float(self.path_list.index(p))/float(len(self.path_list))))
                tags = collect_dicomdir_info(p,tempdir=self.tempdir)
                if not tags is None:
                    for row in tags:
                        if not row['state']=='ERROR':
                            self.tag_list.append(row)
                        else:
                            errors.append(row)
            self.progress(0.)
        # This should sort the list into your initial order for the tree - maybe implement a more customised sort if necessary?
        from operator import itemgetter
        self.sorted_list = sorted(self.tag_list, key=itemgetter('name','date','time','studyuid','series','seriesuid','instance','instanceuid'))

        if len(errors)>0:
            logpath = os.path.join(self.tempdir,'readerrors_log.txt')
            with open(logpath,'w') as outfile:
                logtext = []
                for error in errors:
                    logtext.append('-------------------------------\n{}\n{}'.format(error['path'],error['exception']))
                outfile.write('\n'.join(logtext))
            msg = "Some files could not be read. Please check the logfile at:\n{}".format(logpath)
            tkinter.messagebox.showwarning("WARNING",msg)

        # Uncomment this block to enable mippydb objects in image directory
        #~ if self.ask_recursive:
            #~ self.mippydbpath = os.path.join(self.dicomdir,"mippydb_r")
        #~ else:
            #~ self.mippydbpath = os.path.join(self.dicomdir,"mippydb")
        #~ with open(self.mippydbpath,'wb') as fileobj:
            #~ pickle.dump(self.sorted_list,fileobj,-1)

        return

    def build_dicom_tree(self):
        #~ print "function_started"
        #~ i=0
        # print(self.dirframe.dicomtree.get_children())
        try:
            for item in self.dirframe.dicomtree.get_children():
                self.dirframe.dicomtree.delete(item)
            print("Existing tree cleared")
        except Exception:
            print("New tree created")
            pass
        repeats_found = False
        n_repeats = 0
        uid_list = []
        to_remove = []
        for scan in self.sorted_list:
            #~ print "Adding to tree: "+scan['path']
            if not self.dirframe.dicomtree.exists(scan['studyuid']):
                #~ i+=1
                self.dirframe.dicomtree.insert('','end',scan['studyuid'],text='------',
                                                values=(scan['date'],scan['name'],scan['studydesc']))
            if not self.dirframe.dicomtree.exists(scan['seriesuid']):
                self.dirframe.dicomtree.insert(scan['studyuid'],'end',scan['seriesuid'],
                                            text='Series '+str(scan['series']).zfill(3),
                                            values=('','',scan['seriesdesc']))
            uid = scan['studyuid']+scan['seriesuid']+scan['instanceuid']
            if not uid in uid_list:
                uid_list.append(uid)
                self.dirframe.dicomtree.insert(scan['seriesuid'],'end',scan['instanceuid'],
                                        text=str(scan['instance']).zfill(3),
                                        values=('','',''))
            else:
                repeats_found = True
                n_repeats+=1
                to_remove.append(scan)
        for scan in to_remove:
            # EXPERIMENTAL - Remove repeated instance from list
            # print("REMOVED: {}".format(scan['seriesdesc']))
            self.sorted_list.remove(scan)

        if repeats_found:
            tkinter.messagebox.showwarning("WARNING",str(n_repeats)+" repeat image UID's found and ignored.")
        self.dirframe.dicomtree.update()
        # print("UIDS: {}".format(len(uid_list)))
        # print("IMS: {}".format(len(self.sorted_list)))

        # Run garbage collect to clear anything left in memory unnecessarily
        gc.collect()

        # Save DICOM tree as a snapshot to be opened again at a later time


        #~ self.master.progress = 100
        return

    def select_modules_directory(self):
        print("Load module directory")


        # for line in sys.path:
        #     print(line)
        # for line in sys.modules:
        #     print(line)

        # Grab existing module directory to put back if dialog is cancelled
        prev_mod_dir = self.moduledir
        prev_eggpath = self.module_eggs

        self.moduledir = None
        self.moduledir = tkinter.filedialog.askdirectory(parent=self,initialdir=self.root_dir,title="Select module directory")
        if not self.moduledir:
            self.moduledir = prev_mod_dir
            return
#        sys.path.append(self.moduledir)
        if prev_mod_dir in sys.path:
            sys.path = [value for value in sys.path if value!=prev_mod_dir]
            # sys.path.remove(prev_mod_dir)
        for eggpath in prev_eggpath:
            if eggpath in sys.path:
                sys.path = [value for value in sys.path if value!=eggpath]
                # sys.path.remove(eggpath)
        self.scan_modules_directory()
        return

    def scan_modules_directory(self,skip_update=False):
        # Remove old module eggs from sys.path
        # try:
        #     to_remove = []
        #     for pathname in sys.path:
        #         if os.path.split(pathname)[1] in self.module_list:
        #             to_remove.append(pathname)
        #         if os.oath.split(pathname)[1] in self.module_eggs:
        #             to_remove.append(pathname)
        #     for pathname in to_remove:
        #         print("Removing {} from sys.path".format(pathname))
        #         sys.path.pop(pathname)
        # except AttributeError:
        #     print("First module list loading - if it isn't, there is a problem")


        importlib.invalidate_caches()

        # for mod in sys.modules:
        #     print(mod)

        self.module_list = []
        self.module_eggs = []
        self.hidden_modules = []
        viewerconfigpath = resource_filename('mippy.mviewer','config')
        with open(viewerconfigpath,'rb') as file_object:
            module_info = pickle.load(file_object)
        module_info['dirname']='mippy.mviewer'
        module_info['version']=''
        module_info['eggpath']=None
        self.module_list.append(module_info)
        modules_to_backup = []

        # dcmmodconfigpath = resource_filename('mippy.mdcmmod','config')
        # with open(dcmmodconfigpath,'rb') as file_object:
        #     module_info = pickle.load(file_object)
        # module_info['dirname']='mippy.mdcmmod'
        # module_info['version']=''
        # module_info['eggpath']=None
        # self.module_list.append(module_info)
        # modules_to_backup = []


        if not (self.moduledir is None or not self.moduledir):
            for folder in os.listdir(self.moduledir):
                if not os.path.isdir(os.path.join(self.moduledir,folder)):
                    # Might be an egg.  Try and read as an egg...
                    if folder.upper().endswith('.EGG'):
                        print("{}: It's an egg!!".format(folder))

                        # Get the properties of the egg
                        this_eggpath = os.path.join(self.moduledir,folder)
                        with ZipFile(this_eggpath, 'r') as modulezip:
                            dirs = []
                            pkg_info = {}
                            with modulezip.open('EGG-INFO/PKG-INFO', 'r') as pkg_info_file:
                                for row in pkg_info_file.readlines():
                                    info = row.decode('utf-8').split(':',1)
                                    if len(info)==2:
                                        pkg_info[info[0]]=info[1].rstrip().lstrip()
                                # print(pkg_info)
                                # print(pkg_info['Name'],pkg_info['Version'])
                                # Try to check for newer version of the module
                                try:
                                    # print(pkg_info)
                                    # print('URL: {}'.format(pkg_info['Home-page']))
                                    if 'api.bintray.com' in pkg_info['Home-page'] and not skip_update:
                                        # Check if proxy required, and if proxy requires authentication
                                        proxies = None
                                        if len(self.proxy_list)>0:
                                            auth_required = False
                                            for proxy in self.proxy_list:
                                                if proxy['auth_required']:
                                                    auth_required = True
                                            # If auth required, collect user/password
                                            from easygui import multpasswordbox
                                            fieldNames = ['Username','Password']
                                            title = 'Proxy authentication required'
                                            fieldValues = [self.user.lower(),'']
                                            message = ('Please enter your username and password for the proxy server to check for '
                                                        'updates to {}'.format(pkg_info['Name']))
                                            fieldValues = multpasswordbox(message,title,fieldNames, fieldValues)

                                            # make sure that none of the fields was left blank
                                            while 1:
                                                if fieldValues == None: break
                                                errmsg = ""
                                                for i in range(len(fieldNames)):
                                                    if fieldValues[i].strip() == "":
                                                        errmsg = errmsg + ('"%s" is a required field.\n\n' % fieldNames[i])
                                                if errmsg == "": break # no problems found
                                                fieldValues = multpasswordbox(errmsg, title, fieldNames, fieldValues)
                                            if not fieldValues is None:
                                                proxies = {}
                                                for proxy in self.proxy_list:
                                                    proxy_info = '{}://{}:{}@{}:{}'.format(     proxy['protocol'],
                                                                                                fieldValues[0],
                                                                                                fieldValues[1],
                                                                                                proxy['address'],
                                                                                                proxy['port'])
                                                    print(proxy_info)
                                                    proxies[proxy['protocol']]=proxy_info
                                        r = requests.get(pkg_info['Home-page'],proxies=proxies)
                                        print(r.json())
                                        latest_version = r.json()['latest_version']
                                        remote_pkg_name = r.json()['name']
                                        print(folder.replace(pkg_info['Version'],latest_version))
                                        print("Installed version: {}\nAvailable version: {}".format(pkg_info['Version'],latest_version))
                                        if parse_version(latest_version)>parse_version(pkg_info['Version']):
                                            download = tkinter.messagebox.askyesno(title="New version available",
                                                                message="A newer version of {} is available. Download?".format(remote_pkg_name))
                                            if download==True:
                                                # Get file info for latest version
                                                r = requests.get(pkg_info['Home-page']+'/versions/'+latest_version+'/files',proxies=proxies)
                                                r_owner=r.json()[0]['owner']
                                                r_repo=r.json()[0]['repo']
                                                r_path=r.json()[0]['path']
                                                r_url = 'https://dl.bintray.com/'+r_owner+'/'+r_repo+'/'+r_path
                                                print(r_url)
                                                r_file = r = requests.get(r_url,proxies=proxies)
                                                # Write r_file.content (a byte stream) to the modules folder and move current folder to backups??
                                                print(r_file)
                                                with open(os.path.join(self.moduledir,r_path),'wb') as r_outfile:
                                                    r_outfile.write(r_file.content)
                                                modules_to_backup.append(this_eggpath)
                                        elif parse_version(latest_version)<parse_version(pkg_info['Version']):
                                            print("Local version is newer! Weird...")

                                except requests.exceptions.ConnectionError:
                                    print("Unable to connect to repository for version check")



                                with modulezip.open('EGG-INFO/top_level.txt', 'r') as toplevel:
                                    for row in toplevel.readlines():
                                        dirs.append(row.decode('utf-8').rstrip().lstrip())
                                    for zipdir in dirs:
                                        # Whether config exists or not, need to remove any reference
                                        # to previously loaded modules of these names
                                        to_remove = []
                                        for mod in sys.modules:
                                            if zipdir in mod:
                                                # print("Existing module found! {}".format(mod))
                                                to_remove.append(mod)
                                        for mod in to_remove:
                                            del(sys.modules[mod])
                                        gc.collect()
                                        cfg_file = zipdir+'/config'
                                        if cfg_file in modulezip.namelist():
                                            # print(zipdir, "exists")
                                            if not this_eggpath in self.module_eggs:
                                                self.module_eggs.append(this_eggpath)
#                                            if not this_eggpath in sys.path:
#                                                sys.path.append(this_eggpath)
#                                                print(sys.path[-1])
                                            with modulezip.open(cfg_file,'r') as file_object:
                                                module_info = pickle.load(file_object)
                                                # module_info['dirname']=pkg_info['Name']+'.'+module_info['dirname']
                                                module_info['version'] = pkg_info['Version']
                                                module_info['eggpath'] = this_eggpath

                                                ## CHECK HERE FOR VERSION CLASH
                                                ##
                                                ##
                                                for mod in self.module_list:
                                                    if (module_info['dirname']==mod['dirname'] and
                                                        module_info['version']==mod['version']):
                                                        print("MODULE CLASH: {} {}".format(module_info['dirname'],module_info['version']))
                                                        self.module_info = []
                                                        return


                                                self.module_list.append(module_info)
                                        else:
                                            # Might be a hidden module used by visible ones. Check for __init__.py
                                            # If found, add directory to hidden imports so they can be cleared when
                                            # module directory is scanned next
                                            if zipdir+'/__init__.py' in modulezip.namelist():
                                                self.hidden_modules.append(zipdir)
                        continue
                    else:
                        continue
                file_list = os.listdir(os.path.join(self.moduledir,folder))
                # Whether config exists or not, need to remove any reference
                # to previously loaded modules of these names
                to_remove = []
                for mod in sys.modules:
                    if str(mod).startswith(folder):
                        # print("Existing module found! {}".format(mod))
                        to_remove.append(mod)
                for mod in to_remove:
                    del(sys.modules[mod])
                    gc.collect()
                if (('__init__.py' in file_list or '__init__.pyc' in file_list)
                    and ('module_main.py' in file_list or 'module_main.pyc' in file_list)
                    and 'config' in file_list):
                    gc.collect()
                    cfg_file = os.path.join(self.moduledir,folder,'config')
                    with open(cfg_file,'rb') as file_object:
                        module_info = pickle.load(file_object)
                        module_info['version']='--uncontrolled--'
                        module_info['eggpath']=None
                        ## CHECK HERE FOR VERSION CLASH
                        ##
                        ##
                        for mod in self.module_list:
                            if (module_info['dirname']==mod['dirname'] and
                                module_info['version']==mod['version']):
                                print("MODULE CLASH: {} {}".format(module_info['dirname'],module_info['version']))
                                self.module_list = []
                                return

                        self.module_list.append(module_info)
            from operator import itemgetter
            self.module_list = sorted(self.module_list,key=itemgetter('version'),reverse=True)
            self.module_list = sorted(self.module_list,key=itemgetter('name'))

        try:
            for item in self.moduleframe.moduletree.get_children():
                self.moduleframe.moduletree.delete(item)
            # print("Existing module tree cleared")
        except Exception:
            # print("New module tree created")
            pass
        for module in self.module_list:
            self.moduleframe.moduletree.insert('','end',module['dirname']+'^'+module['version'],
                text=module['name'],values=(module['description'],module['author'],module['version']))

        if len(modules_to_backup)>0:
            if not os.path.exists(os.path.join(self.moduledir,'prev_versions')):
                os.makedirs(os.path.join(self.moduledir,'prev_versions'))
            for mod in modules_to_backup:
                os.rename(mod,os.path.join(self.moduledir,'prev_versions',os.path.split(mod)[1]))
            self.scan_modules_directory(skip_update=True)

        #~ self.master.progress = 50.

#        for mod in sys.modules:
#            print(mod)
        return

    def exit_program(self):
        self.asktoexit()
        return

    def load_docs(self):
        print("Open documentation")
        webbrowser.open_new(MIPPYDOCS)
        return

    def report_issue(self):
        print("Report issue")
        #~ tkinter.messagebox.showinfo("Issue reporting",'Please include the title of your issue in the subject, and a description in the body of the email.\n\n'+
                            #~ 'Where possible, please attach the appropriate log file from MIPPY/logs. Log files are date/time stamped.')
        webbrowser.open_new('https://tree.taiga.io/project/robflintham-mippy/issues')
        return

    def display_version_info(self):
        print("Display version info")
        info = ""
##        verpath = resource_filename('mippy','resources/version.info')
##        with open(verpath,'rb') as infofile:
##            info = infofile.read()
        #~ import mippy
        #~ version = mippy.__version__
        #~ if 'b' in version:
            #~ testing = '(BETA TESTING VERSION)'
        #~ elif 'a' in version:
            #~ testing = '(ALPHA TESTING VERSION)'
        #~ elif 'rc' in version:
            #~ testing = '(Release candidate version)'
        #~ else:
            #~ testing=''
        #~ info = 'Version '+version+'\n'+testing
        from subprocess import Popen,PIPE,check_output
        try:
            output = check_output(['pip','show','mippy'])
        except FileNotFoundError:
            try:
                output = check_output(['pip3','show','mippy'])
            except:
                raise
        except:
            raise
        info = output
        tkinter.messagebox.showinfo("MIPPY: Version info",info)
        return

    def display_changelog(self):
        """
        This has been removed for the time being.
        """
        print("Display changelog")
        info = ""
        with open('docs/changelog.info','rb') as infofile:
            info = infofile.read()
        info_view = Toplevel(self.master)
        info_view.text = Text(info_view,width=120,height=30)
        info_view.text.insert(END,info)
        info_view.text.config(state='disabled')
        info_view.text.see('1.0')
        info_view.text.pack()

    def load_selected_module(self):
        # Remove any existing/leftover paths from sys.path
        if self.moduledir in sys.path:
            sys.path = [value for value in sys.path if value!=self.moduledir]
            # sys.path.remove(self.moduledir)
        for eggpath in self.module_eggs:
            if eggpath in sys.path:
                sys.path = [value for value in sys.path if value!=eggpath]
                # sys.path.remove(eggpath)
        importlib.invalidate_caches()

        eggpath = None

        # Run a garbage collect to clear anything left over from previous module loading
        gc.collect()

        try:

            selected_module = self.moduleframe.moduletree.selection()[0]
            moduledir = selected_module.split('^')[0]
            name = self.moduleframe.moduletree.item(selected_module)['text']
            version = self.moduleframe.moduletree.item(selected_module)['values'][2]
            print(name,version)
            module_name = moduledir+'.module_main'
            for mod in self.module_list:
                if moduledir==mod['dirname'] and version==mod['version']:
                    # Update sys.path with the correct values
                    # print("Found module in module list")
                    if not mod['eggpath'] is None:
                        eggpath = mod['eggpath']
                        if not mod['eggpath'] in sys.path:
                            sys.path.append(mod['eggpath'])
                    else:
                        if not self.moduledir in sys.path:
                            sys.path.append(self.moduledir)
                    # print(sys.path)
            sys.path_importer_cache = {}

            # Need to remove any reference
            # to previously loaded modules of this name
            to_remove = []
            names = [mod['dirname'] for mod in self.module_list]+self.hidden_modules
            # print(names)
            for mod in sys.modules:
                for name in names:
                    if str(mod).startswith(name):
                        to_remove.append(mod)
                # if moduledir in mod:
                #     to_remove.append(mod)
                # if name in mod:
                #     to_remove.append(mod)
            # print(to_remove)
            for mod in to_remove:
                # print("Removing reference to old module {}".format(mod))
                try:
                    del(sys.modules[mod])
                except KeyError:
                    # Assume already deleted
                    pass

            # for path in sys.path:
            #     print(path)
            # print("Loader:\n{}".format(importlib.find_loader(module_name)))
            active_module = importlib.import_module(module_name)
            imp.reload(active_module)
#            if not module_name in sys.modules:
#                active_module = importlib.import_module(module_name)
#            else:
#                active_module = importlib.import_module(module_name)
#                imp.reload(active_module)
            preload_dicom = active_module.preload_dicom()
            try:
                flatten_list = active_module.flatten_series()
            except:
                message = ("\n\n======================================\n"+
                        "   !!!! MIPPY HAS BEEN UPDATED !!!!\n"+
                        "======================================\n\n"+
                        "Please add a function to your module\n"+
                        "as follows:\n\n"+
                        "def flatten_series():\n"+
                        "    return False\n\n"+
                        "or return True if you require all\n"+
                        "images in a single 1D list.\n"+
                        "======================================\n\n")
                print(message)
                flatten_list = True
            if preload_dicom:
                # Attempted to make this section discrete function for use in modules etc
                self.datasets_to_pass = load_images_from_uids(self.sorted_list,self.active_uids,self.tempdir,self.multiprocess)


            else:
                self.datasets_to_pass = []
                previous_tag = None
                for tag in self.sorted_list:
                    if previous_tag:
                        if tag['seriesuid']==previous_tag['seriesuid']:
                            new_series = False
                        else:
                            new_series = True
                    else:
                        new_series = True
                    if tag['instanceuid'] in self.active_uids:
                        if not tag['path'] in self.datasets_to_pass:
                            if new_series:
                                self.datasets_to_pass.append([tag['path']])
                            else:
                                self.datasets_to_pass[-1].append(tag['path'])
                        previous_tag = tag

                #~ gc.collect()
            #~ gc.collect()
            if flatten_list:
                self.datasets_to_pass = list(itertools.chain.from_iterable(self.datasets_to_pass))

            # Generate an instance ID for the module, and write useful information to a temp file
            # that the module can call on
            import datetime
            now = datetime.datetime.now()
            modstamp = name+'_'+version+'_{:02d}{:02d}{:02d}{:02d}{:02d}{:02d}{:06d}'.format(now.year % 1000,
                                        now.month,
                                        now.day,
                                        now.hour,
                                        now.minute,
                                        now.second,
                                        now.microsecond
                          )
            import hashlib
            _hash = hashlib.new('md5')
            _hash.update(bytes(modstamp,'utf-8'))
            instance_id = str(_hash.hexdigest()[0:16]).upper()

            # Grab image paths to pass to module
            image_paths = []
            for tag in self.sorted_list:
                if tag['instanceuid'] in self.active_uids:
                    image_paths.append(tag['path'])


            instance_info = {
                'module_name': self.moduleframe.moduletree.item(selected_module)['text'],
                'module_version': self.moduleframe.moduletree.item(selected_module)['values'][2],
                'module_instance': instance_id,
                'image_directory': self.dicomdir,
                'user_directory': self.userdir,
                'temp_directory': self.tempdir,
                'mippy_version': self.mippy_version,
                'user': self.user,
                'image_paths': image_paths,
                'module_file': active_module.__file__,
                'egg_path': eggpath,
                'module_dirname': moduledir
                }
            # print(instance_info)

            active_module.execute(self.master,instance_info,self.datasets_to_pass)
        except:
            raise
            print("Did you select a module?")
            print("Bet you didn't.")
        return



    def clear_temp_dir(self):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)

    def view_header(self):
        if not hasattr(self, 'active_uids'):
            tkinter.messagebox.showerror('ERROR','No image selected.')
            return
        if len(self.active_uids)>1:
            tkinter.messagebox.showerror('ERROR','You can only view header for a single image/slice at a time.')
            return
        if len(self.active_uids)<1:
            tkinter.messagebox.showerror('ERROR','No image selected.')
            return
        for tag in self.sorted_list:
            if tag['instanceuid'] in self.active_uids:
                dcm_view = Toplevel(self.master)
                dcm_view.text = Text(dcm_view,width=120,height=30)
                ds = pydicom.dcmread(tag['path'])
                if 'SpectroscopyData' in dir(ds):
                    ds.SpectroscopyData=0
                dcm_view.text.insert(END,str(ds))
                dcm_view.text.config(state='disabled')
                dcm_view.text.see('1.0')
                dcm_view.text.pack()
        pass

    def compare_headers(self):
        if not hasattr(self, 'active_uids'):
            tkinter.messagebox.showerror('ERROR','No image selected.')
            return
        if len(self.active_uids)<1:
            tkinter.messagebox.showerror('ERROR','No image selected.')
            return
        if not len(self.active_uids)==2:
            tkinter.messagebox.showerror('ERROR','You can only compare headers for 2 single images/slices at a time.')
            return
        dicoms = []
        for tag in self.sorted_list:
            if tag['instanceuid'] in self.active_uids:
                dicoms.append(pydicom.dcmread(tag['path']))
        for ds in dicoms:
            if 'SpectroscopyData' in dir(ds):
                ds.SpectroscopyData = 0
        diffs = compare_dicom(*dicoms)
        dcm_compare = Toplevel(self.master)
        dcm_compare.text = Text(dcm_compare,width=120,height=30)
        dcm_compare.text.tag_config('highlight', foreground='red')
        dcm_compare.text.tag_config('unhighlight', foreground='black')
        if len(diffs)==0:
            dcm_compare.text.insert(END,'No differences found.')
        else:
            dcm_compare.text.insert(END,'DIFFERENCES IN DICOM HEADER (Some tags ignored)\n')
            for row in diffs:
                #~ print row
                dcm_compare.text.insert(END,'\n'+row[0]+':\n')
                dcm_compare.text.insert(END,'1: '+row[1]+'\n','highlight')
                dcm_compare.text.insert(END,'2: '+row[2]+'\n','highlight')
        dcm_compare.text.config(state='disabled')
        dcm_compare.text.pack()

    def export_dicom(self):

        #~ outdir = os.path.join(self.root_dir,"EXPORT")
        if not hasattr(self, 'active_uids'):
            tkinter.messagebox.showerror('ERROR','No images selected.')
            return
        if len(self.active_uids)<1:
            tkinter.messagebox.showerror('ERROR','No images selected.')
            return
        self.exportdir = tkinter.filedialog.askdirectory(parent=self,initialdir=r"M:",title="Select export directory")
        if self.exportdir is None:
            return
        i=0
        for tag in self.sorted_list:
            if tag['instanceuid'] in self.active_uids:
                fileio.export_dicom_file(load_images_from_uids([tag],self.active_uids,self.tempdir,multiprocess=False)[0][0],self.exportdir)
                i+=1
                self.progress(float(i)/float(len(self.active_uids))*100.)
        self.progress(0.)
        tkinter.messagebox.showinfo('EXPORT FINISHED','Images have finished exporting to:\n'+self.exportdir)
        self.exportdir = None
        return

    def show_log(self):
        logwin = Toplevel()
        logtext = ScrolledText.ScrolledText(logwin)
        with open(self.logpath,'rb') as logfile:
            text = logfile.readlines()
            for row in text:
                logtext.insert(END,row+'\n')
        logtext.pack()

    def enable_multiprocessing(self):
        if self.frozen:
            tkinter.messagebox.showerror('ERROR','Multiprocessing can only be enabled when running the development version of MIPPY.')
            return
        self.multiprocess = True
        tkinter.messagebox.showinfo("INFO","Multiprocessing enabled")
        return

    def disable_multiprocessing(self):
        self.multiprocess = False
        tkinter.messagebox.showinfo("INFO","Multiprocessing disabled")
        return

#########################################################



from mippy.mdicom.siemens import get_ascconv
import pydicom
import numpy as np
from nibabel.nicom import csareader as csar
import json

def get_ascii_value(dicom_ds,searchterm):
    if not ascii in dir(dicom_ds):
        asc = get_ascconv(dicom_ds)
        setattr(dicom_ds,'ascii',asc)
    else:
        print("Using saved ASCII")
    return_value = None
    #~ header = self.get_ascconv()
    found = False
    for s in dicom_ds.ascii:
        if searchterm in s[0]:
            return_value = s[1]
            found = True
            break
    if not found:
        return None
    else:
        try:
            return_value = float(return_value)
        except:
            pass
    return return_value

def is_number(s):
    """
    Tests if a string is readable as a number
    """
    try:
        float(s)
    except ValueError:
        return False
    return True

def get_serial_number(dicom_ds):
    return str(dicom_ds.DeviceSerialNumber)

def get_patient_name(dicom_ds):
    return str(dicom_ds.PatientName)

def get_image_orientation(dicom_ds):
    orient = dicom_ds.ImageOrientationPatient
    # Check TRA
    if ( abs(orient[0])>abs(orient[1]) and abs(orient[0])>abs(orient[2])
        and abs(orient[4])>abs(orient[3]) and abs(orient[4])>abs(orient[5])):
            return "TRA"
    # Check SAG
    elif ( abs(orient[1])>abs(orient[0]) and abs(orient[1])>abs(orient[2])
        and abs(orient[5])>abs(orient[3]) and abs(orient[5])>abs(orient[4])):
            return "SAG"
    # Check COR
    elif ( abs(orient[0])>abs(orient[1]) and abs(orient[0])>abs(orient[2])
        and abs(orient[5])>abs(orient[3]) and abs(orient[5])>abs(orient[4])):
            return "COR"
    else:
        print("ORIENTATION NOT DETECTED", orient)
        return "UNKNOWN"

def get_sequence_type(dicom_ds):
    try:
        seq = dicom_ds.SequenceName
    except:
        # Quick and dirty exception for newer Siemens enhanced format
        seq = dicom_ds[0x21,0x1177].value
    if "se2d" in seq:
        return "SE 2D"
    elif "epfid2d" in seq:
        return "EPI GRE 2D"
    elif "ep2d_diff" in seq:
        return "EPI DIFF 2D"
    elif "ep2d_se" in seq:
        return "EPI SE 2D"
    elif "fl2d" in seq:
        return "GRE 2D"
    elif "fl3d" in seq:
        return "GRE 3D"
    elif "tse2d" in seq:
        return "TSE 2D"
    elif "tse3d" in seq:
        return "TSE 3D"
    elif "tfl2d" in seq:
        return "TGRE 2D"
    elif "tfl3d" in seq:
        return "TGRE 3D"
    else:
        print("SEQUENCE NOT DETECTED", seq)
        return "UNKNOWN"

def get_fov(dicom_ds):
    try:
        fov = str(dicom_ds[0x51,0x100c].value)
        fov = fov.split(" ")[1]
        fov = fov.split("*")
        return ",".join(str(dim) for dim in fov)
    except:
        pxspc_x = dicom_ds.PixelSpacing[0]
        pxspc_y = dicom_ds.PixelSpacing[1]
        rows = dicom_ds.Rows
        cols = dicom_ds.Columns
        fov = str(int(np.round(pxspc_x*cols,0)))+","+str(int(np.round(pxspc_y*rows,0)))
        return fov

def get_acq_matrix(dicom_ds):
    try:
        matrix = dicom_ds.AcquisitionMatrix
    except AttributeError:
        # AcquisitionMatrix tag doesn't exist. Need to
        # work it out from other tags.
        phase = dicom_ds.InPlanePhaseEncodingDirection
        freq_steps = dicom_ds[0x18,0x9058].value
        phase_steps = dicom_ds[0x18,0x9231].value
        # phase_3d = self.dicom[0x18,0x9232].value
        if phase=='ROW':
            matrix = str(phase_steps)+','+str(freq_steps)
            return matrix
        elif phase=='COLUMN':
            matrix = str(freq_steps)+','+str(phase_steps)
            return matrix

    # This needs explanation!
    # You get a 4-long list of values, in this order.
    # [ FREQ-ROWS, FREQ-COLS, PHASE-ROWS, PHASE-COLS ]
    # So if you phase encode by column, you'll get
    # [ VALUE, zero, zero, VALUE ]
    # And if you phase encode by row, you'll get
    # [ zero, VALUE, VALUE, zero ]
    if not matrix[0]==0:
        # Column phase encoding
        # Row in [0], Column in [3]
        matrix = str(matrix[0])+','+str(matrix[3])
        return matrix
    else:
        # Row phase encoding
        # Row in [2], Column in [1]
        matrix = str(matrix[2])+','+str(matrix[1])
        return matrix

def get_recon_matrix(dicom_ds):
    rows = dicom_ds.Rows
    cols = dicom_ds.Columns
    return str(cols)+','+str(rows)

def get_acq_slice_thickness(dicom_ds):
    if dicom_ds.MRAcquisitionType=='2D':
        return float(dicom_ds.SliceThickness)
    elif dicom_ds.MRAcquisitionType=='3D':
        try:
            slice_resolution = dicom_ds[0x19,0x1017].value
            slice_thickness = dicom_ds.SliceThickness
            if slice_resolution<1:
                acq_slice_thickness = slice_thickness / slice_resolution
                return float(acq_slice_thickness)
            else:
                return float(slice_thickness)
        except KeyError:
            print("Cannot read slice resolution tag")
            return "UNKNOWN"
    else:
        print("Do not understand acquisition acquisition type")
        return "UNKNOWN"

def get_oversampling(dicom_ds):
    oversampling = get_ascii_value(dicom_ds,'sKSpace.dPhaseOversamplingForDialog')

    if not oversampling is None:
        return float(oversampling)
    else:
        # Assume no oversampling if tag not written
        return 0.

def get_recon_slice_thickness(dicom_ds):
    return float(dicom_ds.SliceThickness)

def get_phase_encode_direction(dicom_ds):
    orient = get_image_orientation(dicom_ds)
    phase = dicom_ds.InPlanePhaseEncodingDirection
    if orient=='TRA':
        if 'ROW' in phase:
            return 'RL'
        elif 'COL' in phase:
            return 'AP'
        else:
            print("Do not understand phase encode direction")
            return "UNKNOWN"
    elif orient=='SAG':
        if 'ROW' in phase:
            return 'AP'
        elif 'COL' in phase:
            return 'FH'
        else:
            print("Do not understand phase encode direction")
            return "UNKNOWN"
    elif orient=='COR':
        if 'ROW' in phase:
            return 'RL'
        elif 'COL' in phase:
            return 'FH'
        else:
            print("Do not understand phase encode direction")
            return "UNKNOWN"
    else:
        print("Do not understand orientation")
        return "UNKNOWN"

def get_nsa(dicom_ds):
    return float(dicom_ds.NumberOfAverages)

def get_TR(dicom_ds):
    return float(dicom_ds.RepetitionTime)

def get_TE(dicom_ds):
    try:
        return np.round(float(dicom_ds.EchoTime),3)
    except:
        # Use effective echo time (copied from from per-frame group to create single slice)
        return np.round(dicom_ds.EffectiveEchoTime,3)

def get_TI(dicom_ds):
    try:
        return float(dicom_ds.InversionTime)
    except AttributeError:
#            print("No inversion time found")
        return "None"

def get_flip_angle(dicom_ds):
    return float(dicom_ds.FlipAngle)

def get_pat_type(dicom_ds):
#        print("PAT NOT YET SUPPORTED")
    return "UNKNOWN"

def get_pat_2d(dicom_ds):
#        print("PAT NOT YET SUPPORTED")
    return "UNKNOWN"

def get_pat_3d(dicom_ds):
#        print("PAT NOT YET SUPPORTED")
    return "UNKNOWN"

def get_partial_fourier_phase(dicom_ds):
#        print("PARTIAL FOURIER NOT YET SUPPORTED")
    return "UNKNOWN"

def get_partial_fourier_slice(dicom_ds):
#        print("PARTIAL FOURIER NOT YET SUPPORTED")
    return "UNKNOWN"

def get_coil_elements(dicom_ds):
    try:
        coilstring = str(dicom_ds[0x51,0x100f].value)
    except KeyError:
        # Attempt alternative DICOM tag for newer scanners
        try:
            coilstring = str(dicom_ds[0x21,0x114f].value)
        except KeyError:
            # Still didn't work
            print("Unable to get coil information")
            return "UNKNOWN"

    coils = read_coil_string(coilstring)
    return ';'.join(coils)



# Some comprehension of coil string required.
# Assume something of the fashion C:HE1-4;SP3;SP5
def read_coil_string(coilstring):
    coils = coilstring.replace('b\'C:','').replace('C:','').replace('\'','').strip().split(';')
    active_coils = []
    for coil in coils:
        if '-' in coil:
            # Convert range into list of individual elements
            # Get prefix
            prefix = ''
            for char in coil:
                if not is_number(char):
                    prefix=prefix+char
                else:
                    break
            # Remove prefx and split at dash
            coil_range = coil.replace(prefix,'').split('-')
            for i in range(int(coil_range[0]),int(coil_range[1])+1):
                active_coils.append(prefix+str(i))
        elif ',' in coil:
            # Convert range into list of individual elements
            # Get prefix
            prefix = ''
            for char in coil:
                if not is_number(char):
                    prefix=prefix+char
                else:
                    break
            # Remove prefix and split at comma
            coil_range = coil.replace(prefix,'').split(',')
            for num in coil_range:
                active_coils.append(prefix+num)
        else:
            active_coils.append(coil)
    return active_coils

def get_bandwidth(dicom_ds):
    return float(dicom_ds.PixelBandwidth)

def get_image_filter(dicom_ds):
    imtype = dicom_ds.ImageType
    if 'FILT' in imtype:
        return "ImageFilter"
    else:
        return 'None'

def get_uniformity_correction(dicom_ds):
    imtype = dicom_ds.ImageType
    if 'NORM' in imtype:
        this_image_norm = True
    else:
        this_image_norm = False
    prescan = get_ascii_value(dicom_ds,'sPreScanNormalizeFilter.ucOn')
    norm_filt = get_ascii_value(dicom_ds,'sNormalizeFilter.ucOn')
    if prescan is None:
        prescan_on = False
    elif prescan == '0x1':
        prescan_on = True
    else:
        # Assume if tag exists, it's on.
        prescan = True
    if norm_filt is None:
        norm_on = False
    elif norm_filt == '0x1':
        norm_on = True
    else:
        # Assume if tag exists, it's on
        norm_filt = True
    if prescan_on and this_image_norm:
        return 'PreScanNormalize'
    elif prescan_on and not this_image_norm:
        return 'PreScanNormalize-Unfiltered'
    elif norm_on and this_image_norm:
        return 'Normalize'
    elif norm_on and not this_image_norm:
        return 'Normalize-Unfiltered'
    else:
        return 'None'

def get_distortion_correction(dicom_ds):
    imtype = dicom_ds.ImageType
    if 'DIS2D' in imtype:
        this_image_dc = '2D'
    elif 'DIS3D' in imtype:
        this_image_dc = '3D'
    else:
        this_image_dc = False

    try:
        # Fudge for XA data
        if dicom_ds[0x8,0x9206].value=='DISTORTED':
            this_image_dc = False
        else:
            this_image_dc = True
    except:
        pass

    dc_filter = get_ascii_value(dicom_ds,'sDistortionCorrFilter.ucMode')
    if dc_filter == '0x2':
        dc_filter_type = '2D'
    elif dc_filter == '0x4':
        dc_filter_type = '3D'
    elif dc_filter == 2:
        dc_filter_type = '2D'
    elif dc_filter == 3 or dc_filter == 4:
        dc_filter_type = '3D'
    else:
        dc_filter_type = False

    if dc_filter_type and this_image_dc:
        return dc_filter_type
    elif dc_filter_type and not this_image_dc:
        return dc_filter_type+'-Unfiltered'
    else:
        return "None"

def get_study_uid(dicom_ds):
    return str(dicom_ds.StudyInstanceUID)

def get_series_number(dicom_ds):
    return dicom_ds.SeriesNumber

def get_series_uid(dicom_ds):
    return str(dicom_ds.SeriesInstanceUID)

def get_instance_number(dicom_ds):
    return dicom_ds.InstanceNumber

def get_sop_uid(dicom_ds):
    return str(dicom_ds.SOPInstanceUID)

def get_study_date(dicom_ds):
    return str(dicom_ds.StudyDate)

def get_study_time(dicom_ds):
    return str(dicom_ds.StudyTime)

def get_institution_name(dicom_ds):
    try:
        return str(dicom_ds.InstitutionName)
    except:
        return "UNKNOWN"

def get_scanner_model(dicom_ds):
    return str(dicom_ds.Manufacturer)+' '+str(dicom_ds.ManufacturerModelName)

def get_institution_address(dicom_ds):
    return str(dicom_ds.InstitutionAddress)

def get_department(dicom_ds):
    try:
        return str(dicom_ds.InstitutionalDepartmentName)
    except AttributeError:
        return "UNKNOWN"

def get_station_name(dicom_ds):
    return str(dicom_ds.StationName)

def get_field_strength(dicom_ds):
    return float(dicom_ds.MagneticFieldStrength)



import dill as pickle
import os
import sys
import numpy as np
import pydicom
from pydicom.dataset import FileDataset, Dataset
import pydicom.uid
from pydicom.uid import generate_uid, PYDICOM_IMPLEMENTATION_UID, ImplicitVRLittleEndian
import tempfile
import datetime

def save_temp_ds(ds,tempdir,fname):
        '''
        I don't know why this bit needs to be done, but if you don't create these strings
        for each slice, python isn't able to pickle the objects properly and complains
        about not having the Attribute _character_set - perhaps the character set isn't
        defined until a string representation of the object is required?
        '''
        ds_str = str(ds)
        if not os.path.exists(tempdir):
                os.makedirs(tempdir)
        temppath = os.path.join(tempdir,fname)
        #~ if not os.path.exists(temppath):
        with open(temppath,'wb') as tempfile:
                pickle.dump(ds,tempfile,protocol=3)
                #pickle.dump(ds,tempfile)
                tempfile.close()
        return

def add_all(dataset1,dataset2):
        groups_to_copy = [int(0x8),int(0x10)]
        for element in dataset2:
                #~ print element.tag.group
                if element.tag.group in groups_to_copy:
                        val = None
                        try:
                                val = dataset1[pydicom.tag.Tag(element.tag)].value
                        except KeyError:
                                dataset1.add_new(element.tag, element.VR, element.value)
                        except:
                                raise
                        if val=="":
                                dataset1.add_new(element.tag, element.VR, element.value)
        return dataset1

def save_dicom(images,directory,
				ref=None,
				series_number='add_thousand' ,
				series_description = "MIPPY saved images",
				series_description_append = None,
				path_append = None,
				fnames = None,
				rescale_slope = 'use_bitdepth',
				rescale_intercept = 'use_bitdepth',
				sop_class = 'use_ref',
				slice_positions = 'use_ref'):

	"""
	Takes a list of numpy arrays of pixels, and either generates
	whole new DICOM objects with minimal information or uses
	reference DICOM objects to generate series description etc
	"""

	if ref is None:
		print("Please pass reference images - ref image list must be the same length as images to be saved")
		pass

	suffix = '.dcm'
	ser_uid = generate_uid()



	if isinstance(images, list):
		# Convert to 3D numpy array for simplicity!
		images = np.array(images)

	images_out = []
	for i in range(len(images)):
		tempname = tempfile.NamedTemporaryFile(suffix=suffix).name
		file_meta = Dataset()
		file_meta.MediaStorageSOPClassUID = ref[i].file_meta.MediaStorageSOPClassUID
		file_meta.ImplementationClassUID = PYDICOM_IMPLEMENTATION_UID
		file_meta.MediaStorageSOPInstanceUID = generate_uid()
		file_meta.TransferSyntaxUID = ImplicitVRLittleEndian

		ds = FileDataset(tempname,{},file_meta=file_meta,preamble=b"\0"*128)
		ds.is_little_endian = True
		ds.is_implicit_VR = True
		add_all(ds,ref[i])
		ds.SeriesInstanceUID = ser_uid
		ds.SOPInstanceUID = generate_uid()
		ds.StudyInstanceUID = ref[i].StudyInstanceUID

		if series_description_append is None:
			ds.SeriesDescription = series_description
		else:
			ds.SeriesDescription = ds.SeriesDescription+series_description_append

		ds.SamplesPerPixel = 1
		ds.PhotometricInterpretation = 'MONOCHROME2'
		ds.Rows = np.shape(images)[1]
		ds.Columns = np.shape(images)[2]
		ds.WindowCenter = np.min(images)+(np.max(images)-np.min(images))/2
		ds.WindowWidth = np.max(images)-np.min(images)
		ds.BitsAllocated = 16
		ds.BitsStored = 12
		ds.InstanceNumber = i
		ds.ImagePositionPatient = ref[i].ImagePositionPatient
		ds.ImageOrientationPatient = ref[i].ImageOrientationPatient
		ds.PixelSpacing = ref[i].PixelSpacing
		ds.HighBit = 11
		ds.FrameOfReferenceUID = ref[i].FrameOfReferenceUID
		ds.PixelRepresentation = 0

		if rescale_slope=='use_ref':
			ds.RescaleSlope = ref[i].RescaleSlope
		elif rescale_slope=='use_bitdepth':
			max = np.max(images)
			min = np.min(images)
			ds.RescaleSlope = (max-min) / (2**ds.BitsStored-1)
		elif not rescale_slope is None:
			try:
				ds.RescaleSlope = float(rescale_slope)
			except:
				ds.RescaleSlope = 1
		else:
			ds.RescaleSlope = 1

		if rescale_intercept=='use_ref':
			ds.RescaleIntercept = ref[i].RescaleIntercept
		elif rescale_intercept=='use_bitdepth':
			ds.RescaleIntercept = 0.+np.min(images)
		elif not rescale_intercept is None:
			try:
				ds.RescaleIntercept = float(rescale_intercept)
			except:
				ds.RescaleIntercept = 0
		else:
			ds.RescaleIntercept = 0

		#~ print(np.max(images[i]),ds.RescaleSlope,ds.RescaleIntercept)

		if series_number=='same':
			ds.SeriesNumber = ref[i].SeriesNumber
		elif series_number=='add_thousand':
			ds.SeriesNumber = ref[i].SeriesNumber+1000
		else:
			ds.SeriesNumber=0
		if not path_append is None:
			outdir = directory+'_'+str(path_append)
		else:
			outdir = directory
		if not os.path.exists(outdir):
			os.makedirs(outdir)
		ds.PixelData = ((images[i]-ds.RescaleIntercept)/ds.RescaleSlope).astype(np.uint16)
		ds.SmallestImagePixelValue = np.min(np.array(ds.PixelData)).astype(np.uint16)
		ds.LargestImagePixelValue = np.max(np.array(ds.PixelData)).astype(np.uint16)
		ds.save_as(os.path.join(outdir,fnames[i]))
	return


# TO finish!


# def load_4d_data(dcm_paths,temp_data_path = None):
#     """
#     Takes a list of paths to DICOM files, and attempts to open as a 4D dataset
#     with consistent X,Y,Z dimensions.
#
#     Allows you to specify a temp_data file holding a pickled dump of a 4D
#     numpy array.
#     """
#     if not temp_data_path is None:
#         "Loading earlier dump of data"
#         win.Im4Dfull = np.load(win.tempdatafile)
#         win.dynamics = np.shape(win.Im4Dfull)[0]
#         win.slices = np.shape(win.Im4Dfull)[1]
#         win.rows = np.shape(win.Im4Dfull)[2]
#         win.columns = np.shape(win.Im4Dfull)[3]
#         win.shape3d = (win.slices,win.rows,win.columns)
#         win.tfull = np.load(win.temptimefile)
#     else:
#         win.Im4Dfull = None
#
#     if win.Im4Dfull is None:
#         slice_positions = []
#         s1 = None
#         s2 = None
#         if len(np.unique(np.array(win.images)))==1:
#             # Same path for every image, most likely single multi-frame image
#             imfiles = [win.images[0]]
#             output(win,"{} image file received...".format(len(imfiles)))
#         else:
#             imfiles = win.images
#             output(win,"{} images received...".format(len(imfiles)))
#         px_list = list(range(len(imfiles)))
#         output(win,"Reading the image data; this may take some time")
#         time.sleep(2)
#         time.sleep(2)
#         i=0
#         win.repetition_time = 0.
#         win.mosaic = False
#         rows = 0
#         columns = 0
#         win.n_in_mosaic = 0
#         win.n_in_mosaic_sq = 0
#         sq = 0
#         classic_dicom = False
#         enhanced = False
#         win.missing_data = False
#         for im in imfiles:
#             i+=1
#             if i%100==0:
#                 output(win,"Reading {}/{}".format(i,len(imfiles)))
#                 gc.collect()
#             #~ output(win,'... opening DICOM')
#             ds = pydicom.dcmread(im,force=True)
#             #~ output(win,'... finished opening DICOM')
#
#             instance = ds.InstanceNumber
#             win.series_number = ds.SeriesNumber
#             try:
#                 win.series_description = ds.SeriesDescription
#             except AttibuteError:
#                 win.series_description = "UNKNOWN SERIES DESCRIPTION"
#
#             if i==1:
#
#                 if 'MOSAIC' in ds.ImageType:                        # Find out how to reshape pixel data if mosaic
#                     win.mosaic = True
#                     rows = ds.Rows
#                     columns = ds.Columns
#                     win.n_in_mosaic = ds[0x19,0x100a].value
#                     found_square = False
#                     while not found_square:
#                         sq+=1
#                         if sq**2>win.n_in_mosaic:
#                             found_square = True
#                     win.n_in_mosaic_sq = sq**2
#                     px_list = list(range(len(imfiles)*win.n_in_mosaic_sq))
#
#                     win.repetition_time = float(ds.RepetitionTime)
#                     win.echo_time = float(ds.EchoTime)
#                     win.PixelSpacing = ds.PixelSpacing
#                     win.slice_thickness = ds.SliceThickness
#                     #~ print(len(px_list))
#                 elif 'ENHANCED' in ds.SOPClassUID.name.upper():
#                     enhanced = True
#                     # Multi-frame image.  Use n_frames as number of images
#                     px_list = list(range(ds.NumberOfFrames))
#                     s1 = ds[0x5200,0x9230][0][0x20,0x9113][0].ImagePositionPatient
#                     s2 = ds[0x5200,0x9230][1][0x20,0x9113][0].ImagePositionPatient
#                     win.repetition_time = float(ds[0x5200,0x9229][0][0x18,0x9112][0].RepetitionTime)
#                     win.echo_time = float(ds[0x5200,0x9230][0][0x18,0x9114][0][0x18,0x9082].value)
#                     win.PixelSpacing = ds[0x5200,0x9230][0][0x28,0x9110][0].PixelSpacing
#                     win.slice_thickness = ds[0x5200,0x9230][0][0x28,0x9110][0].SliceThickness
#                 else:
#                     classic_dicom = True
#                     s1 = ds.ImagePositionPatient
#                     win.repetition_time = float(ds.RepetitionTime)
#                     win.echo_time = float(ds.EchoTime)
#                     win.PixelSpacing = ds.PixelSpacing
#                     win.slice_thickness = ds.SliceThickness
#             elif i==2 and classic_dicom:
#                 s2 = ds.ImagePositionPatient
#             if not win.mosaic and not enhanced:
#                 # e.g. Siemens data with a single slice per image
#                 max_instance = 0
#                 win.missing_data = False
#                 try:
#                     px_list[instance-1] = ds.pixel_array.astype(np.float64)     # Need to use instance to place slices in list, as can't guarantee files are read in the correct order
#                 except IndexError:
#                     # Missing data from start of series - attempt to correct by placing in "instance - n_images" and then rolling array afterwards by required amount
#                     px_list[instance-1-len(imfiles)] = ds.pixel_array.astype(np.float64)
#                     if instance>max_instance:
#                         max_instance = instance
#                     win.missing_data = True
#                 pos = ds.ImagePositionPatient
#                 if not pos in slice_positions:
#                     slice_positions.append(pos)
#             elif enhanced:
#                 # Lots of frames per file, but only a single slice per frame
#                 # Potential hole for Enhanced, assuming only a single file loaded
#                 for n in range(ds.NumberOfFrames):
#                     if n%100==0:
#                         output(win,"Reading frame {}/{}".format(n,ds.NumberOfFrames))
#                     px_list[n] = get_px_array(ds,enhanced,n+1)
#                     pos = ds[0x5200,0x9230][n][0x20,0x9113][0].ImagePositionPatient
#                     if not pos in slice_positions:
#                         slice_positions.append(pos)
#             else:
#                 # Mosaic image - all slices in a single image, 1 image per dynamic
#                 px = ds.pixel_array.astype(np.float64)
#                 #~ print range(sq)
#                 for y in range(sq):
#                         for x in range(sq):
#                             #~ print x,y
#                             imrows = rows//sq
#                             imcols = columns//sq
#                             px_list[(instance-1)*sq*sq+y*sq+x]=np.array(px[y*imrows:(y+1)*imrows,x*imcols:(x+1)*imcols])
#                 pos = ds.ImagePositionPatient
#                 if not pos in slice_positions:
#                     slice_positions.append(pos)
#         del(ds)
#         gc.collect()
#         if win.missing_data:
#             output(win,'Missing instances from start of series - rearranging data as necessary')
#             for n in range(max_instance-len(px_list)):
#                 px_list.append(px_list.pop(0))
#         Im3D = np.array(px_list)
#         del(px_list)
#         gc.collect()
#         print('Im3D shape:',np.shape(Im3D))
#
#         if win.mosaic:
#             sort_order = 'dsrc'
#         elif s1==s2:
#             sort_order='sdrc'
#         else:
#             sort_order='dsrc'
#
#         if not win.mosaic:
#             win.frames = np.shape(Im3D)[0]
#             win.rows = np.shape(Im3D)[1]
#             win.columns = np.shape(Im3D)[2]
#             win.slices = len(slice_positions)
#             win.dynamics = int(win.frames/win.slices)   #~ added by ND 09/05/2018 after error when running - something like "dynamics not found..."
#             print(win.dynamics,win.slices,win.rows,win.columns)
#         else:
#             win.frames = np.shape(Im3D)[0]
#             win.slices = int(win.n_in_mosaic_sq)
#             win.dynamics = int(win.frames/win.slices)
#             win.rows = int(rows/np.sqrt(win.n_in_mosaic_sq))
#             win.columns = int(columns/np.sqrt(win.n_in_mosaic_sq))
#             print('MOSAIC DETECTED:',win.dynamics,win.slices,win.rows,win.columns)
#         win.shape3d = (win.slices,win.rows,win.columns)
#
#         if sort_order=='drcs':
#             win.Im4Dfull = Im3D.reshape(win.dynamics,win.rows,win.columns,win.slices).swapaxes(1,2).swapaxes(1,3)
#         elif sort_order=='sdrc':
#             win.Im4Dfull = Im3D.reshape(win.slices,win.dynamics,win.rows,win.columns).swapaxes(0,1)
#         elif sort_order=='dsrc':
#             win.Im4Dfull = Im3D.reshape(win.dynamics,win.slices,win.rows,win.columns)
#         del(Im3D)
#         gc.collect()
#         output(win,'4D array shape: {},{},{},{}'.format(win.dynamics,win.slices,win.rows,win.columns))
#         #~ print dynamics,slices,rows,columns
#         win.tfull = (np.array(range(win.dynamics)).astype(np.float64))*win.repetition_time/1000
#
#         #~ np.savetxt(os.path.join(outdir,"Im4D.txt"),Im4Dfull)
#
#         #~ win.Im4Dfull.dump(win.tempdatafile)
#         #~ win.tfull.dump(win.temptimefile)
#
#     # Load pixel data to canvas
#     # Load average across all timepoints for each slice
#     # Attempting this in a one-liner using list comprehension and an averaging along first axis
#     win.im1.load_images([imslice for imslice in np.mean(win.Im4Dfull,axis=0)])
#     output(win,'Average for all dynamics displayed on canvas')
#
#
#
#     for i in range(win.slices):
#         win.iminfo[0].append(collect_acq_info(win.images[i]))
#     return



from easygui import diropenbox,choicebox,msgbox
import sys
import os
from subprocess import call
import getpass

user = getpass.getuser()

DEFAULT_DIR = r'C:\MIPPY_'+str(user)

MESSAGE = (
	'This sets up a MIPPY working directory and creates a shortcut\n'
	+'in the Start Menu from which you can launch MIPPY.\n\n'
	+'The working directory is where you should install\n'
	+'your modules for MIPPY.\n\n'
	+'Would you like to use the default directory ('+DEFAULT_DIR+') or\n'
	+'select a different directory?')

choices = ['Default directory ('+DEFAULT_DIR+')','Different direcrtory']

user_choice = choicebox(msg=MESSAGE,title="INSTALL MIPPY",
				choices=choices,
				preselect=0,run=True)

if user_choice==choices[1]:
	install_dir = diropenbox("Select a directory in which to install MIPPY",
						"SELECT INSTALL DIRECTORY",
						DEFAULT_DIR)
elif user_choice==choices[0]:
	install_dir = DEFAULT_DIR
else:
	sys.exit()

#~ print(install_dir)
module_dir = os.path.join(install_dir,'modules')

if not os.path.exists(module_dir):
	os.makedirs(module_dir)
os.environ['MIPPYPATH']='\"'+install_dir+'\"'

start_command = '@echo off\nmippy \"'+install_dir+'\"'
#~ print(start_command)

with open(os.path.join(install_dir,'start_mippy.bat'),'w') as f:
	f.write(start_command)

# Create Start Menu folder:
start_menu_dir=os.path.join(os.environ['userprofile'],'Start Menu','Programs','MIPPY')
if not os.path.exists(start_menu_dir):
	os.makedirs(start_menu_dir)

mippy_link_path = os.path.join(start_menu_dir,'MIPPY.lnk')

if os.path.exists(mippy_link_path):
	os.remove(mippy_link_path)

command=('powershell $s=(New-Object -COM WScript.Shell).CreateShortcut(\''
			+os.path.join(start_menu_dir,'MIPPY.lnk')
			+'\');$s.TargetPath=\''
			+os.path.join(install_dir,'start_mippy.bat')
			+'\';$s.Save()')
#~ print(command)
call(command)



if os.path.exists(mippy_link_path):
	END_MESSAGE = (
		'MIPPY was successfully installed!\n\n'
		+'You can now run MIPPY from the Start Menu, or from\n'
		+'the selected installation directory of:\n\n'
		+install_dir
		)
	end = msgbox(msg=END_MESSAGE,title="Success!")
else:
	END_MESSAGE = (
		'Unable to create shortcut on Start Menu.  Please\n'
		+'seek help.\n\n'
		+'In the meantime, you should still be able to run\n'
		+'MIPPY from your selected installation directory of:\n\n'
		+install_dir
		)
	end = msgbox(msg=END_MESSAGE,title="Oops...")


# Generic functions for MIPPY windows that don't fit anywhere else

def optionmenu_patch(om, var):
        menu = om['menu']
        last = menu.index("end")
        for i in range(0, last+1):
                menu.entryconfig(i, variable=var)

from collections import Mapping, Container
from sys import getsizeof

def deep_getsizeof(o, ids): 
        # From https://code.tutsplus.com/tutorials/understand-how-much-memory-your-python-objects-use--cms-25609
        """Find the memory footprint of a Python object

        This is a recursive function that drills down a Python object graph
        like a dictionary holding nested dictionaries with lists of lists
        and tuples and sets.

        The sys.getsizeof function does a shallow size of only. It counts each
        object inside a container as pointer only regardless of how big it
        really is.

        :param o: the object
        :param ids:
        :return:
        """
        d = deep_getsizeof
        if id(o) in ids:
                return 0

        r = getsizeof(o)
        ids.add(id(o))

        if isinstance(o, str) or isinstance(0, str):
                return r

        if isinstance(o, Mapping):
                return r + sum(d(k, ids) + d(v, ids) for k, v in o.items())

        if isinstance(o, Container):
                return r + sum(d(x, ids) for x in o)

        return r
        
def getsizeof(an_object):
        from pympler import asizeof
        return asizeof.asizeof(an_object)


# -*- coding: utf-8 -*-
"""
This code is heavily based on code found online.
-- Rob
"""

   # <-- Python 2.5 ONLY
import tkinter.tix
import time
from PIL import Image, ImageTk

class SplashScreen( object ):
   def __init__( self, tkRoot, imageFilename, minSplashTime=0 ):
      self._root              = tkRoot
      self._imopen            = Image.open(imageFilename)
      self._image             = ImageTk.PhotoImage( self._imopen )
      self._splash            = None
      self._minSplashTime     = time.time() + minSplashTime
      
   def __enter__( self ):
      # Remove the app window from the display
      self._root.withdraw( )
      
      # Calculate the geometry to center the splash image
      scrnWt = self._root.winfo_screenwidth( )
      scrnHt = self._root.winfo_screenheight( )
      
      imgWt = self._image.width()
      imgHt = self._image.height()
      
      imgXPos = (scrnWt / 2) - (imgWt / 2)
      imgYPos = (scrnHt / 2) - (imgHt / 2)

      # Create the splash screen      
      self._splash = tkinter.tix.Toplevel()
      self._splash.overrideredirect(1)
      self._splash.geometry( '+%d+%d' % (imgXPos, imgYPos) )
      tkinter.tix.Label( self._splash, image=self._image, cursor='watch' ).pack( )

      # Force Tk to draw the splash screen outside of mainloop()
      self._splash.update( )
   
   def __exit__( self, exc_type, exc_value, traceback ):
      # Make sure the minimum splash time has elapsed
      timeNow = time.time()
      if timeNow < self._minSplashTime:
         time.sleep( self._minSplashTime - timeNow )
      
      # Destroy the splash window
      self._splash.destroy( )
      
      # Display the application window
      self._root.deiconify( )


#--------------------------------------------
# Now putting up splash screens is simple



# Create the tkRoot window
#tkRoot = Tix.Tk( )
#
#with SplashScreen( tkRoot, 'splashImage.jpg', 3.0 ):
#   initializeMyApplication( )
#   buildTheGUI( tkRoot )
#
#tkRoot.mainloop( )


from tkinter import *
from tkinter.ttk import *
from tkinterhtml import HtmlFrame
import markdown2
import os
import webbrowser
import shutil


class MarkdownWindow(Toplevel):
    """
    Window to render simple HTML for displaying simple text-only markdown
    as rendered HTML.
    """
    def __init__(self,master_window,inline='',filepath=''):
        if not inline=='' and not filepath=='':
            print("ERROR: You cannot pass both inline content and filepath.\n"
                    +"Please pass only one or the other.")
        if inline=='' and filepath=='':
            print("ERROR: No content provided.\n"
                    +"Please pass either inline (string) content or filepath.")
        Toplevel.__init__(self,master=master_window)
        self.protocol("WM_DELETE_WINDOW", self.on_close)

        self.htmlframe = HtmlFrame(self,horizontal_scrollbar = "auto")
        self.htmlframe.set_content('<html></html>')
        md_content = None
        if not inline=='':
            md_content = (markdown2.markdown(inline))
        else:
            md_content = markdown2.markdown_path(filepath)
        print(md_content)
        self.htmlframe.set_content(md_content)
        self.htmlframe.pack()

        return

    def on_close(self):
        print("Closed help window, event captured")
        self.master.lift()
        self.destroy()
        return

def markdown_to_browser(module_window,inline='',filepath=''):
    """
    Takes markdown input (as a filepth or as a string), renders
    to html and displays in a web browser. Images must be in a subdirectory
    called img and must be referenced as 'img/filename' in the markdown text.

    Must be called from a module in order to get appropriate directory structure
    for the temporary files.
    """
    md_content = None
    img_paths = []
    if not inline=='':
        md_content = (markdown2.markdown(inline))
    else:
        md_content = markdown2.markdown_path(filepath)
        # Check for images
        filedir = os.path.split(filepath)[0]
        if os.path.exists(os.path.join(filedir,'img')):
            for img in os.listdir(os.path.join(filedir,'img')):
                img_paths.append(os.path.join(filedir,'img',img))
    # print(md_content)
    htmlpath = os.path.join(module_window.instance_info['temp_directory'],module_window.instance_info['module_instance'])
    if not os.path.exists(htmlpath):
        os.makedirs(htmlpath)
    with open(os.path.join(htmlpath,'index.html'),'w') as html_file:
        html_file.write(md_content)
    if len(img_paths)>0:
        if not os.path.exists(os.path.join(htmlpath,'img')):
            os.makedirs(os.path.join(htmlpath,'img'))
        for path in img_paths:
            indir = os.path.split(path)[0]
            infile = os.path.split(path)[1]
            shutil.copyfile(path,os.path.join(htmlpath,'img',infile))
    webbrowser.open_new(os.path.join(htmlpath,'index.html'))
    return



"""
XML generators and readers for use with pydicom objects.

This is to improve serialisation and (potentially) reduce the file-size
of MIPPY temp files.  Pixel data is dumped separately with numpy.

The file_meta, header and pixel data are stored as 3 separate files
(where [UID] is the SOP instance UID).

[UID]_meta.xml
[UID]_header.xml
[UID]_px.pix
"""


from datetime import datetime
import os
from . import logging
from . import logging.handlers
import sys

class RedirectText(object):
        def __init__(self, log):
                self.logfile = log
                self.write("LOG FILE: "+str(datetime.now())+'\n')
        def write(self, stringobj):
                with open(self.logfile,'a') as f:
                        f.write(stringobj)

class StdOutCatcher(object):
        def __init__(self,logobject):
                self.log = logobject
        def write(self,s):
                self.log.info(s)

class StdErrCatcher(object):
        def __init__(self,logobject):
                self.log = logobject
        def write(self,s):
                self.log.exception(s)

def setup_logging():
        # Code for logging from https://www.loggly.com/ultimate-guide/python-logging-basics/
        # with addition of custom classes StdErrCatcher and StdOutCatcher
        starttime = str(datetime.now())
        logdir = os.path.join(os.getcwd(),"MIPPY-logs")
        try:
                os.makedirs(logdir)
        except:
                pass
        
        logpath=os.path.join(logdir,starttime.replace(":",".").replace(" ","_")+".txt")
        
        with open(logpath,'wb') as logfile:
                logfile.write("MIPPY LOG: "+starttime+'\n')
        
        handler = logging.handlers.WatchedFileHandler(
                os.environ.get("LOGFILE",logpath)
                )
        formatter = logging.Formatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        logout = logging.getLogger()
        logout.setLevel(os.environ.get("LOGLEVEL","INFO"))
        logout.addHandler(handler)
        logout.info("Is this working?")
        
        sys.stderr = StdErrCatcher(logout)
        sys.stdout = StdOutCatcher(logout)
        
        logout.info("How about now?")
        return


from tkinter import *
from tkinter.ttk import *
import tkinter.messagebox
from mippy.viewing import *
import os
from PIL import Image,ImageTk
import platform
from pkg_resources import resource_filename

def preload_dicom():
        """
        This method is essential for the module to run. MIPPY needs to know whether
        the module wants preloaded DICOM datasets or just paths to the files.  Most
        modules will probably want preloaded DICOM datasets so that you don't have
        to worry about different manufacturers, enhanced vs non-enhanced etc.
        However, I imagine some people will prefer to work with the raw data files
        and open the datasets themselves?
        """
        # If you want DICOM datasets pre-loaded, return True.
        # If you want paths to the files only, return False.
        # Note the capital letters on True and False.  These are important.
        return True

def flatten_series():
        """
        By default, MIPPY will pass the image datasets as a 2D list, divided by series.
        If you want only a single, flat list from all your series, return True.
        """
        return True


def execute(master_window,instance_info,images):
        print("Module loaded...")
        print("Received "+str(len(images))+" image datasets.")
        print(os.getcwd())
        #~ icondir = os.path.join(os.getcwd(),'source','images')


        # This is to fix backwards compatibility for anywhere the old dicomdir object
        # was used
        dicomdir = instance_info['image_directory']

        window = Toplevel(master_window)
        window.instance_info = instance_info
        window.title("{} {}: {}".format(window.instance_info['module_name'],window.instance_info['mippy_version'],window.instance_info['module_instance']))
        # Create canvas
        if platform.system()=='Linux':
                canvas_size=320
        else:
                canvas_size=512
        window.imcanvas = MIPPYCanvas(window,bd=0,width=canvas_size,height=canvas_size,drawing_enabled=True)
        window.imcanvas.enable_advanced_rois()
        # Open icons for button
        sq_im = resource_filename('mippy','resources/square_roi.png')
        el_im = resource_filename('mippy','resources/ellipse_roi.png')
        fr_im = resource_filename('mippy','resources/freehand_roi.png')
        li_im = resource_filename('mippy','resources/line_roi.png')
        window.roi_sq_im = ImageTk.PhotoImage(file=sq_im)
        window.roi_el_im = ImageTk.PhotoImage(file=el_im)
        window.roi_fr_im = ImageTk.PhotoImage(file=fr_im)
        window.roi_li_im = ImageTk.PhotoImage(file=li_im)

        window.toolbar=Frame(window)

        window.roi_square_button = Button(window.toolbar,text="Draw square ROI",command=lambda:window.imcanvas.draw_rectangle_roi(),image=window.roi_sq_im)
        window.roi_ellipse_button = Button(window.toolbar,text="Draw elliptical ROI",command=lambda:window.imcanvas.draw_ellipse_roi(),image=window.roi_el_im)
        window.roi_polygon_button = Button(window.toolbar,text="Draw freehand ROI", command=lambda:window.imcanvas.draw_freehand_roi(),image=window.roi_fr_im)
        window.roi_line_button = Button(window.toolbar,text="Draw line",command=lambda:window.imcanvas.draw_line_roi(),image=window.roi_li_im)
        window.roi_square_button.config(pad=2)
        window.roi_ellipse_button.config(pad=2)
        window.roi_polygon_button.config(pad=2)
        window.roi_line_button.config(pad=2)
        #~ window.scrollbutton = Button(window, text="SLICE + / -")
        window.imcanvas.img_scrollbar = Scrollbar(window,orient='horizontal')
        window.imcanvas.configure_scrollbar()
        window.statsbutton = Button(window,text="Get ROI statistics",command=lambda:get_stats(window))
        window.profilebutton = Button(window,text="Get Horizontal Profile",command=lambda:get_profile(window))
        window.vertprofilebutton = Button(window,text="Get Vertical Profile",command=lambda:get_vert_profile(window))
        window.statstext = StringVar()
        window.statswindow = Label(window,textvariable=window.statstext)
        window.zoominbutton = Button(window,text="ZOOM +",command=lambda:zoom_in(window))
        window.zoomoutbutton = Button(window,text="ZOOM -",command=lambda:zoom_out(window))

        # Pack GUI using "grid" layout

        window.roi_square_button.grid(row=0,column=0)
        window.roi_ellipse_button.grid(row=1,column=0)
        window.roi_polygon_button.grid(row=2,column=0)
        window.roi_line_button.grid(row=3,column=0)
        window.imcanvas.grid(row=0,column=0,columnspan=1,rowspan=6,sticky='nsew')
        window.toolbar.grid(row=0,column=1,sticky='nsew')
        #~ window.scrollbutton.grid(row=7,column=0,sticky='nsew')
        window.imcanvas.img_scrollbar.grid(row=6,column=0,sticky='ew')
        window.statsbutton.grid(row=3,column=1,sticky='ew')
        window.profilebutton.grid(row=4,column=1,sticky='ew')
        window.vertprofilebutton.grid(row=5,column=1,sticky='ew')
        window.statswindow.grid(row=6,column=1,columnspan=1,rowspan=3,sticky='nsew')
        window.zoominbutton.grid(row=7,column=0,sticky='nsew')
        window.zoomoutbutton.grid(row=8,column=0,sticky='nsew')

        window.imcanvas.load_images(images)

        for im in window.imcanvas.images:
            print("RS: {}; RI: {}; SS: {}".format(im.rs, im.ri, im.ss))

        window.rowconfigure(0,weight=1)
        window.columnconfigure(0,weight=1)

        return

def close_window(active_frame):
        active_frame.destroy()
        return

def get_stats(window):
#        canvas = window.imcanvas
#        px = canvas.get_roi_pixels()
#        mean = np.mean(px)
#        std = np.std(px)
#        area = len(px)*canvas.get_active_image().xscale*canvas.get_active_image().yscale
#        tkMessageBox.showinfo('ROI STATS','Mean: %s\nStandard Deviation %s\nArea: %s' %(np.round(mean,2),np.round(std,2),np.round(area,2)))
        results = window.imcanvas.get_roi_statistics()
        tkinter.messagebox.showinfo('ROI STATS',results)

#        display_results(results,window)

def get_profile(window):
#        if not window.imcanvas.roi_list[0].roi_type=='line':
#                print "Only a line ROI can be used to generate profiles"
#                return
        profile,scale = window.imcanvas.get_profile()
        print(profile)
        return
def get_vert_profile(window):
        profile,scale = window.imcanvas.get_profile(direction='vertical')
        print(profile)
        return

def zoom_in(window):
        pass

def zoom_out(window):
        pass



import os
from datetime import datetime
from .mdicom.reading import load_images_from_uids

import shutil
import fnmatch
import stat
import itertools

def list_all_files(dirpath,recursive=False):
        pathlist = []
        if recursive:
                for root,directories,files in os.walk(dirpath):
                        for filename in files:
                                filepath = os.path.join(root,filename)
                                pathlist.append(filepath)
        else:
                allobjects = os.listdir(dirpath)
                for f in allobjects:
                        thispath = os.path.join(dirpath,f)
                        if not os.path.isdir(thispath):
                                pathlist.append(thispath)
        return pathlist

def save_results(results,name=None,directory=None):
        """
        Standardised way of saving results in TXT files. Not sure what
        to do with them afterwards yet...
        """
        timestamp = str(datetime.now()).replace(" ","_").replace(":","")[0:21]
        # Truncates at milisecond level

        if not name:
                fname = "RESULTS_"+timestamp+".txt"
        else:
                fname = name+"_"+timestamp+".txt"

        if directory is None:
                current_dir = os.getcwd()
                outputdir = os.path.join(current_dir,"Results")
                if not os.path.exists(outputdir):
                        os.makedirs(outputdir)
        else:
                outputdir = directory
                if not os.path.exists(outputdir):
                        os.makedirs(outputdir)
        outpath = os.path.join(outputdir,fname)
        with open(outpath,'w') as txtfile:
                txtfile.write(results)
        return

def export_dicom_file(ds,outdir):
        """
        Export a DICOM dataset to the disk drive.
        At some point this will be customisable!
        """
        dir1 = str(ds.PatientName).replace('^','__')+"_"+remove_invalid_characters(ds.PatientID)
        dir2 = ds.StudyDate+"_"+ds.StudyTime
        dir3 = str(ds.SeriesNumber).zfill(4)+"_"+remove_invalid_characters(str(ds.SeriesDescription))

        #~ fname1 = str(ds.ImageType).replace('/','-').strip()+"_"
        fname1 = ''.join(str(i)+"_" for i in ds.ImageType)
        fname2 = str(ds.SeriesNumber).zfill(4)+"_"
        fname3 = str(ds.InstanceNumber).zfill(5)+"_"
        fname4 = str(ds.SOPInstanceUID)[-8:]

        fext = '.DCM'

        outdirfull = os.path.join(outdir,dir1,dir2,dir3)
        if not os.path.exists(outdirfull):
                os.makedirs(outdirfull)

        ds.save_as(os.path.join(outdirfull,fname1+fname2+fname3+fname4+fext))
        return

def remove_invalid_characters(value):
        deletechars = '\\/:*?"<>|'
        for c in deletechars:
                value = value.replace(c,'')
        return value

def copyToDir(src, dst, updateonly=True, symlinks=True, ignore=None, forceupdate=None, dryrun=False):

    def copySymLink(srclink, destlink):
        if os.path.lexists(destlink):
            os.remove(destlink)
        os.symlink(os.readlink(srclink), destlink)
        try:
            st = os.lstat(srclink)
            mode = stat.S_IMODE(st.st_mode)
            os.lchmod(destlink, mode)
        except OSError:
            pass  # lchmod not available
    fc = []
    if not os.path.exists(dst) and not dryrun:
        os.makedirs(dst)
        shutil.copystat(src, dst)
    if ignore is not None:
        ignorepatterns = [os.path.join(src, *x.split('/')) for x in ignore]
    else:
        ignorepatterns = []
    if forceupdate is not None:
        forceupdatepatterns = [os.path.join(src, *x.split('/')) for x in forceupdate]
    else:
        forceupdatepatterns = []
    srclen = len(src)
    for root, dirs, files in os.walk(src):
        fullsrcfiles = [os.path.join(root, x) for x in files]
        t = root[srclen+1:]
        dstroot = os.path.join(dst, t)
        fulldstfiles = [os.path.join(dstroot, x) for x in files]
        excludefiles = list(itertools.chain.from_iterable([fnmatch.filter(fullsrcfiles, pattern) for pattern in ignorepatterns]))
        forceupdatefiles = list(itertools.chain.from_iterable([fnmatch.filter(fullsrcfiles, pattern) for pattern in forceupdatepatterns]))
        for directory in dirs:
            fullsrcdir = os.path.join(src, directory)
            fulldstdir = os.path.join(dstroot, directory)
            if os.path.islink(fullsrcdir):
                if symlinks and dryrun is False:
                    copySymLink(fullsrcdir, fulldstdir)
            else:
                if not os.path.exists(fulldstdir) and dryrun is False:
                    os.makedirs(fulldstdir)
                    shutil.copystat(fullsrcdir, fulldstdir)
        for s,d in zip(fullsrcfiles, fulldstfiles):
            if s not in excludefiles:
                if updateonly:
                    go = False
                    if os.path.isfile(d):
                        srcdate = os.stat(s).st_mtime
                        dstdate = os.stat(d).st_mtime
                        if srcdate > dstdate:
                            go = True
                    else:
                        go = True
                    if s in forceupdatefiles:
                        go = True
                    if go is True:
                        fc.append(d)
                        if not dryrun:
                            if os.path.islink(s) and symlinks is True:
                                copySymLink(s, d)
                            else:
                                shutil.copy2(s, d)
                else:
                    fc.append(d)
                    if not dryrun:
                        if os.path.islink(s) and symlinks is True:
                            copySymLink(s, d)
                        else:
                            shutil.copy2(s, d)
    return fc



import copy
from pydicom.tag import Tag
import pydicom

def get_frame_ds_old(frame,ds):
    # "frame" starts from 1, not 0.  Need to subtract 1 for correct indexing.
    slicenum = frame-1
    n_frames = ds.NumberOfFrames
    print("Extracting frame "+str(frame)+" of "+str(n_frames))
#        print "    Copying original dataset"
    ds_new = copy.deepcopy(ds)
#        ds_new = pickle.loads(pickle.dumps(ds))
#        ds_new = ujson.loads(ujson.dumps(ds))
#        ds_new = Dataset()
#        ds_new = add_all(ds_new,ds)
#        print ds._character_set
#        print "    Stripping out old tags"
    del ds_new[Tag(0x5200,0x9230)]
    del ds_new[Tag(0x5200,0x9229)]
#        print "    Adding new tags"
    ds_new = add_all(ds_new,ds[0x5200,0x9229][0])
    ds_new = add_all(ds_new,ds[0x5200,0x9230][slicenum])
#        print "    Correcting rows and columns"
    rows = int(ds_new.Rows)
    cols = int(ds_new.Columns)
#        print "    Copying pixel data"
    try:
        ds_new.PixelData = ds.PixelData[slicenum*(rows*cols*2):(slicenum+1)*(rows*cols*2)]
    except:
        # Added to skip pixel data when ds is loaded without
        pass
#        print "    Replacing instance number"
    ds_new.InstanceNumber = frame
    ds_new.NumberOfFrames = 1
    # print(ds_new._character_set)
    # for i in range(len(ds_new._character_set)):
    #         ds_new._character_set[i] = pydicom.charset.default_encoding
    # print(ds_new._character_set)
    # ds_new._character_set = ds._character_set
    # ds_new._character_set = pydicom.charset.default_encoding
    # ds_new = ds_new._replace(_character_set=pydicom.charset.default_encoding)
#        print "    Returning split dataset"
    print(ds_new)

    return ds_new

def get_frame_ds(frame,ds):
    """
    This new function creates a DICOM dataset from scratch rather than performing
    a deepcopy, (hopefully) providing a speed increase over the old method.
    """
    slicenum=frame-1    # Frame numbers indexed from 1, not 0
    n_frames = ds.NumberOfFrames
    print("Extracting frame {} of {}".format(frame,n_frames))

    file_meta = ds.file_meta
    filename = ds.filename


    ds_new = pydicom.dataset.FileDataset(filename,{},file_meta=file_meta,preamble=b"\0"*128)
    ds_new.file_meta = ds.file_meta
    ds_new.preamble=bytes("\0"*128,'utf-8')
    if 'Explicit' in ds_new.file_meta.TransferSyntaxUID.name:
        ds_new.is_implicit_VR = False
    else:
        ds_new.is_implicit_VR = True
    if 'Little' in ds_new.file_meta.TransferSyntaxUID.name:
        ds_new.is_little_endian = True
    else:
        ds_new.is_little_endian = False
    # ds_new.is_implicit_VR = True
    # ds_new.is_little_endian = True
    # Add all tags from ds to ds_new
    ds_new = add_all_simple(ds_new,ds)
    # Delete "enhanced" tags
    del ds_new[Tag(0x5200,0x9229)]
    del ds_new[Tag(0x5200,0x9230)]
    # Re-add nested tags from enhanced per-frame groups to main DICOM header
    ds_new = add_all(ds_new,ds[0x5200,0x9229][0])
    ds_new = add_all(ds_new,ds[0x5200,0x9230][slicenum])
    # Correct pixel data to be of the correct frame
    rows = int(ds_new.Rows)
    cols = int(ds_new.Columns)
    # ds_new.PixelData = ds.PixelData[slicenum*(rows*cols*2):(slicenum+1)*(rows*cols*2)]
    if 'PixelData' in dir(ds_new):
        ds_new.PixelData = ds.PixelData[slicenum*(rows*cols*2):(slicenum+1)*(rows*cols*2)]
    # Reset InstanceNumber and number of frames
    ds_new.InstanceNumber = frame
    ds_new.NumberOfFrames = 1
    # Change SOP Class UID to MR Image Storage
    ds_new.SOPClassUID = '1.2.840.10008.5.1.4.1.1.4'
    # Return new dataset
    ds_str = str(ds_new)
    return ds_new

def add_all_simple(dataset1,dataset2):
    for element in dataset2:
        dataset1.add_new(element.tag,element.VR,element.value)
    return dataset1

def add_all(dataset1,dataset2):
        for element in dataset2:
                if element.VR=="SQ":
                        for D in range(len(element.value)):
                                add_all(dataset1,element.value[D])
                else:
                        try:
                                del dataset1[Tag(element.tag)]
                        except KeyError:
                                pass
                        except:
                                raise
                        dataset1.add_new(element.tag, element.VR, element.value)
        return dataset1



from setuptools import setup, find_packages
from codecs import open
from os import path
import sys
import os

here = path.abspath(path.dirname(__file__))




##################################

def get_version():
        import mippy
        return mippy.__version__

def check_version():
        version = get_version()
        print(version)
        import datetime
        import os
        # Get timestamp for __init__.pyc
        version_time = datetime.datetime.fromtimestamp(os.path.getmtime(r'mippy\__pycache__\__init__.cpython-36.pyc'))
        print(version_time)
        other_times = []
        for root, dirs, files in os.walk('mippy'):
                for f in files:
                        fpath = os.path.join(root,f)
                        lastdir = os.path.split(os.path.split(fpath)[0])[1]
                        if os.path.splitext(fpath)[1]=='.py' or lastdir=='resources':
                                other_times.append(datetime.datetime.fromtimestamp(os.path.getmtime(fpath)))
        code_changed = False
        for tstamp in other_times:
##                print tstamp, version_time, tstamp>version_time
                if tstamp>version_time:
                        code_changed = True
                        print(tstamp)
                        break
        if code_changed:
                print("CANNOT COMPILE - VERSION NUMBER OUTDATED")
                import sys
                sys.exit()
        return

# Determine version number from BUILD tags on gitlab
try:
    if os.environ.get('CI_COMMIT_TAG'):
            if os.environ['CI_COMMIT_TAG'].startswith('v'):
                version = os.environ['CI_COMMIT_TAG'][1:]
            else:
                version = os.environ['CI_COMMIT_TAG']
    else:
        version = '0.'+os.environ['CI_JOB_ID'] # Use job ID if no commmit tag provided
except KeyError:
    import datetime
    version='0.'+str(datetime.datetime.now())[0:23].replace(' ','-').replace(':','')

with open('requirements.txt','r') as f:
    requirements = f.readlines()

setup(       name='mippy',
                version=version,
                description='Modular Image Processing in Python',
                author='Robert Flintham',
                author_email='rbf906@gmail.com',
                license='BSD-3-Clause',
                classifiers=[
                        'Programming Language :: Python :: 3',
                        ],
                install_requires=requirements,
                packages=['mippy','mippy.mdicom','mippy.mviewer','mippy.mdicom.miners'],
                scripts=['scripts/mippy.bat','scripts/_mippy.py','scripts/install_mippy.bat','scripts/install_mippy.py'],
                url='https://tree.taiga.io/project/robflintham-mippy/',
                package_data={'':['resources/*','mviewer/config','luts/*']}
        )



import pydicom
from nibabel.nicom import csareader as csar

def get_ascii_header(ds):
        """
        Expects a pydicom dataset.
        Uses nibabel's "nicom" module to extract CSA data
        """
        csa = None
        try:
            csa = csar.get_csa_header(ds,'series')
        except:
            pass
        if not csa is None:
            try:
                    csa_header = csa['tags']['MrPhoenixProtocol']['items'][0]
            except TypeError:
                csa_header = csa['tags']['MrProtocol']['items'][0]

        if csa is None:
            # Fall back to returning direct text read of file
            # This will probably be horrendously slow...
            with open(ds.filename,'r',errors='replace') as original_file:
                csa_header = original_file.read()
        return csa_header

def get_ascconv(ds):
        ascii = get_ascii_header(ds)
        start_string = '### ASCCONV BEGIN'
        end_string = 'ASCCONV END ###'
        ascconv = ascii[ascii.find(start_string):ascii.find(end_string)+len(end_string)].split('\n')
        asc_split = []
        for i in range(len(ascconv)-1):
                if i == 0:
                        continue
                asc_split.append(ascconv[i].replace(' ','').replace('\t','').split('='))
        #~ for i in range(10):
                #~ print asc_split[i]
        return asc_split

def get_ascii_value(searchterm,ds):
    return_value = None
    header = get_ascconv(ds)
    found = False
    for s in header:
        if searchterm in s[0]:
            return_value = s[1]
            found = True
            break
    if not found:
        return None
    # Disabled this - doesn't appear to hold true....?
    #~ if return_value=='0x1':
        #~ return_value = True
    #~ elif return_value=='0x4':
        #~ return_value = False
    else:
        try:
            return_value = float(return_value)
        except:
            pass
    return return_value



#~ from multiprocessing import Pool
from multiprocessing.dummy import Pool as ThreadPool
import os
import pydicom
from . import io
from . import pixel
from pkg_resources import resource_filename
from subprocess import call
import dill as pickle
import numpy as np
import sys
from PIL import Image
import gc

from ..viewing import get_overlay
from ..threading import multithread
from functools import partial
from .mrenhanced import get_frame_ds
import itertools

def recursive_file_finder(path):
        pathlist = []
        for root,directories,files in os.walk(path):
                for filename in files:
                        filepath = os.path.join(root,filename)
                        pathlist.append(filepath)
        return pathlist

def get_instance_uid(dicomfile):
        #~ try:
        ds = pydicom.dcmread(dicomfile)
        io.save_temp_ds(ds,r'K:\MIPPYTEMP',ds.SOPInstanceUID)
        #~ except:
                #~ return None
        return ds.SOPInstanceUID

def multidicomread(paths,threads=None):
        if threads:
                pool = ThreadPool(threads)
        else:
                pool = ThreadPool()
        results = pool.map(get_instance_uid, paths)
        pool.close()
        pool.join()
        return results

def get_dataset(info,tempdir=None):
        """
        This function uses instance UID and/or filepath and/or instance
        number to find a single DICOM dataset. It checks existing MIPPY
        temp files, and if it can't be found, loads the file/instance asked
        for. This allows parallelisation of the process to speed up loading
        of datasets.
        info is a tuple of:
        info[0] = UID (from MIPPY, i.e. including instance number for
                        enhanced DICOM)
        info[1] = absolute file path
        info[2] = instance number

        UID cannot be None. The other two can, but it might not find anything.
        """
        uid = info[0]
        filepath = info[1]
        instance = info[2]

        # First, check for UID in temp files
        if not uid is None:
                temppath = os.path.join(tempdir,str(uid)+'.mds')
                if os.path.exists(temppath):
                        print("TEMP FILE FOUND: "+str(uid))
                        with open(temppath,'rb') as tempfile:
                                ds = pickle.load(tempfile)
                        return ds
                elif not filepath is None:
                        ds = pydicom.dcmread(filepath)
                        if not 'ENHANCED' in ds.SOPClassUID.name.upper():
                                print("SINGLE DICOM FILE FOUND")
                                io.save_temp_ds(ds,tempdir,str(ds.SOPInstanceUID)+'.mds')
                                return ds
                        else:
                                if not instance is None:
                                        from .mrenhanced import get_frame_ds
                                        #~ print("FETCHING SPLIT DATASET")
                                        ds_split = get_frame_ds(instance,ds)
                                        #~ io.save_temp_ds(ds_split,tempdir,str(ds.SOPInstanceUID)+"_"+str(i).zfill(3)+'.mds')
                                        return ds_split

        # If it's got this far, cannot find file or not enough info
        print("CANNOT FIND FILE")
        return None


def collect_dicomdir_info(path,tempdir=None,force_read=False):
    try:
        tags=[]
        ima_mrs_uid = '1.3.12.2.1107.5.9.1'
        # Variables to contain path to dcmdjpeg for compressed DICOM
        dcmdjpeg = None
        # This automatically excludes Philips "XX_" files, but only based on name.  If they've been renamed they
        # won't be picked up until the "try/except" below.
        if os.path.split(path)[1].startswith("XX_"):
                return tags
#        print os.path.split(path)[1]

        # Remove any previous datasets just held as "ds" in memory
        if 'ds' in locals():
                del(ds)
        ds = None
        #Read file, if not DICOM then ignore
        try:
                ds = pydicom.dcmread(path, force=force_read)
                # Removed this garbage collection call to try speed up directory reading
                #~ gc.collect()
        except Exception:
                print(path+'\n...is not a valid DICOM file and is being ignored.')
                return tags
        if ds and not ds is None:
                #~ print path

                try:
                        # There has to be a better way of testing this...?
                        # If "ImageType" tag doesn't exist, then it's probably an annoying "XX" type file from Philips
                        type = ds.ImageType
                except Exception:
                        return tags
                # Ignore "OT" (other?) modality DICOM objects - for now at least...
                # Particularly for Siemens 3D data viewing!
                modality = ds.Modality
                if (
                        'OT' in modality
                        ):
                        return tags

                transfer_syntax =  ds.file_meta[0x2,0x10].value.name
                if 'JPEG' in transfer_syntax:
                        compressed = True
                else:
                        compressed = False
                try:
                        # Some manufacturers use a handy "series description" tag, some don't
                        seriesdesc = ds.SeriesDescription
                except Exception:
                        try:
                                # Some store "protocol name", which will do for now until I find something better
                                seriesdesc = ds.ProtocolName
                        except Exception:
                                # If all else fails, just use a generic string
                                seriesdesc = "Unknown Study Type"

                if "PHOENIXZIPREPORT" in seriesdesc.upper():
                        # Ignore any phoenix zip report files from Siemens
                        return tags
                # Unless told otherwise, assume normal MR image storage
                mode = "Assumed MR Image Storage"
                sop_class_uid = pydicom.uid.UID(ds.SOPClassUID)
                try:
                        mode = sop_class_uid.name
                except:
                        raise
                if "SOFTCOPY" in mode.upper() or "BASIC TEXT" in mode.upper():
                        # Can't remember why I have these, I think they're possible GE type files???
                        return tags
                if mode.upper()=="ENHANCED MR IMAGE STORAGE":
                        # If enhanced file, record number of frames.  This is important for pulling the right imaging
                        # data out for the DICOM tree and image previews
                        enhanced = True
                        frames = ds.NumberOfFrames
                else:
                        enhanced = False
                        frames = 1
                study_uid = str(ds.StudyInstanceUID)
                series_uid = str(ds.SeriesInstanceUID)
                name = str(ds.PatientName)
                date = ds.StudyDate
                series = ds.SeriesNumber
                time = ds.StudyTime
                try:
                        # Some manufacturers use a handy "study description" tag, some don't
                        studydesc = ds.StudyDescription
                except Exception:
                        try:
                                # Philips stores "body part examined", which will do for now until I find something better
                                studydesc = ds.BodyPartExamined
                        except Exception:
                                # If all else fails, just use a generic string
                                studydesc = "Unknown Study Type"



                #~ if tags is None:
                        #~ tags = []

                if enhanced:
                        # Set "instance" array to match number of frames
                        instance = np.array(list(range(frames)))+1
                else:
                        # Or if not enhanced/multi-frame, just create a single element list so that the code
                        # below still works
                        try:
                                instance = [ds.InstanceNumber]
                        except AttributeError:
                                print("INSTANCE NUMBER TAG DOESN'T EXIST")
                                print(path)
                                raise

                instance_uid = 'UNKNOWN'

                original_series_uid = series_uid
                original_series_desc = seriesdesc

                for i in instance:

                        recon = None


                        if not enhanced:
                                instance_uid = str(ds.SOPInstanceUID)


                        else:
                                # Append instance UID with the frame number to give unique reference to each slice
                                instance_uid = str(ds.SOPInstanceUID)+"_"+str(i).zfill(4)

                                # If Real or Imaginary, prepend this to study description and add it to series UID to separate out in
                                # MIPPY browser window
                                try:
                                        if 'REAL' in ds[0x5200,0x9230][i-1][0x18,0x9226][0][0x8,0x9208].value:
                                                recon = 'R'
                                        elif 'IMAGINARY' in ds[0x5200,0x9230][i-1][0x18,0x9226][0][0x8,0x9208].value:
                                                recon = 'I'
                                        elif 'PHASE' in ds[0x5200,0x9230][i-1][0x18,0x9226][0][0x8,0x9208].value:
                                                recon = 'P'
                                        elif 'MAGNITUDE' in ds[0x5200,0x9230][i-1][0x18,0x9226][0][0x8,0x9208].value:
                                                recon = 'M'
                                except KeyError:
                                        pass
                                except:
                                        raise



                        if not recon is None:
                                series_uid = original_series_uid+'_'+recon
                                seriesdesc = recon+': '+original_series_desc

                        if enhanced:
                            series_uid += '_E'
                            seriesdesc = '[E] '+seriesdesc


                        """
                        Disabled to test Python's built-in JPEG module
                        Reenabled!!!
                        """
                        if compressed:
                                # Check if temp file already exists for that InstanceUID. If so, read that file. If not,
                                # uncompress the file and replace ds. Dataset will get saved as temp file at the end of this
                                # function.
                                temppath = os.path.join(tempdir,instance_uid+'.mds')
                                if os.path.exists(temppath):
                                        print(seriesdesc+' '+str(i).zfill(3))
                                        print("    COMPRESSED DICOM - Temp file found")
                                        with open(temppath,'rb') as tempfile:
                                                ds = pickle.load(tempfile)
                                        tempfile.close()
                                else:
                                        if dcmdjpeg is None:
                                                if 'darwin' in sys.platform:
                                                        dcmdjpeg = os.path.join(tempdir,"dcmdjpeg_mac")
                                                elif 'linux' in sys.platform:
                                                        dcmdjpeg = os.path.join(tempdir,"dcmdjpeg_linux")
                                                elif 'win' in sys.platform:
                                                        dcmdjpeg = os.path.join(tempdir,"dcmdjpeg_win.exe")
                                                else:
                                                        print("UNSUPPORTED OPERATING SYSTEM")
                                                        print(str(sys.platform))
                                        # Uncompress the file
                                        outpath=os.path.join(tempdir,'UNCOMP_'+instance_uid+'.DCM')
                                        print(seriesdesc+' '+str(i).zfill(3))
                                        print("    COMPRESSED DICOM - Uncompressing")
                                        command = [dcmdjpeg,'-v',path,outpath]
                                        call(command, shell=False)
                                        del(ds)
                                        ds = pydicom.dcmread(outpath)
                                        # Fix for incorrect JPEG UID in file_meta
                                        if 'JPEG' in ds.file_meta.TransferSyntaxUID.name:
                                                ds.file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'
                                                ds.is_little_endian = True
                                                ds.is_implicit_VR = False

                        print(name,"/",date,"/",seriesdesc,"/",i)
#                        print(mode.upper())
                        if not ("SPECTROSCOPY" in mode.upper() or ima_mrs_uid in mode.upper()):
                                pxfloat=pixel.get_px_array(ds,enhanced,i,bitdepth=8)
                                if pxfloat is None:
                                        print("ERROR: Couldn't read pixel data")
                                        continue
                                try:
                                        overlay = get_overlay(ds)
                                        if not overlay is None:
                                                pxfloat[np.where(overlay>0)]=255
                                        #~ pxfloat += overlay
                                except KeyError:
                                        pass
                                except:
                                        raise
                        else:
                                mrs_image =  resource_filename('mippy','resources/mrs.png')
                                pxfloat = np.asarray(Image.open(mrs_image)).astype(np.float64)
                                pxfloat = np.mean(pxfloat,axis=2)
                                series_uid = series_uid+'.RAW'
                                seriesdesc = '<MRS> '+seriesdesc

                        # Append the information to the "tag list" object
                        tags.append(dict([('date',date),('time',time),('name',name),('studyuid',study_uid),
                                        ('series',series),('seriesuid',series_uid),('studydesc',studydesc),
                                        ('seriesdesc',seriesdesc),('instance',i),('instanceuid',instance_uid),
                                        ('path',path),('enhanced',enhanced),('compressed',compressed),
                                        ('px_array',pxfloat),('state','READ_SUCCESS')]))
                # Assuming all this has worked, serialise the dataset (ds) for later use, with the instance UID
                # as the file name
                if not enhanced:
                        if tempdir:
                                io.save_temp_ds(ds,tempdir,instance_uid+'.mds')

        del ds

        return tags
    except Exception as e:
        return [{'state':'ERROR','path':path,'exception':e}]

def compare_dicom(ds1,ds2,diffs=None,num=None,name=''):
        if diffs is None:
                diffs = []
                #~ gc.collect()
        if num:
                num=' ('+str(num).zfill(4)+')'
        else:
                num=''
        exclude_list = ['UID',
                                'REFERENCE',        # Don't care what localisers were used
                                'SERIES TIME',
                                'ACQUISITION TIME',
                                'CONTENT TIME',
                                'CREATION TIME',
                                'PIXEL VALUE',
                                'WINDOW',
                                'CSA',                # Still don't really know what CSA is
                                'PIXEL DATA',
                                'PADDING',
                                'PHOENIX',
                                'PRIVATE TAG DATA']
        for element in ds1:
                if any(s in element.name.upper() for s in exclude_list):
                        continue
                val1 = element.value
                try:
                        val2 = ds2[element.tag].value
                except:
                        diffs.append((name+str(element.name)+num+' '+str(element.tag.group).zfill(4)+','+str(element.tag.element).zfill(4),str(val1),'--MISSING--'))
                        continue
                if element.VR=="SQ":
                        for i in range(len(val1)):
                                compare_dicom(val1[i],val2[i],diffs=diffs,num=i,name=str(element.name)+' >> ')
                        continue
                elif not val1==val2:
                        if not any(s in element.name.upper() for s in exclude_list):
                                diffs.append((name+str(element.name)+num+' '+str(element.tag.group).zfill(4)+','+str(element.tag.element).zfill(4),str(val1),str(val2)))
        for element in ds2:
                val2 = element.value
                try:
                        val1 = ds1[element.tag].value
                except:
                        diffs.append((name+str(element.name)+num+' '+str(element.tag.group).zfill(4)+','+str(element.tag.element).zfill(4),'--MISSING--',str(val2)))
                        continue

        return diffs

def load_images_from_uids(list_of_tags,uids_to_match,tempdir,multiprocess=False):
        datasets_to_pass = []
        dcm_info = []
        previous_tag = None
        open_file = None
        # Disabled multiprocessing for loading images into modules due to problems with get_dataset function on Python3
        multiprocess = False
        if not multiprocess or ('win' in sys.platform and not 'darwin' in sys.platform and len(uids_to_match)<25):
                for tag in list_of_tags:
                        if tag['instanceuid'] in uids_to_match or tag['seriesuid'] in uids_to_match:
                                # Check to see if new series
                                if previous_tag:
                                        if tag['seriesuid']==previous_tag['seriesuid']:
                                                new_series = False
                                        else:
                                                new_series = True
                                else:
                                        new_series = True
                                # First, check if dataset is already in temp files
                                temppath = os.path.join(tempdir,tag['instanceuid']+'.mds')
                                if os.path.exists(temppath):
                                        #~ print "TEMP FILE FOUND",tag['instanceuid']
                                        with open(temppath,'rb') as tempfile:
                                                if new_series:
                                                        datasets_to_pass.append([pickle.load(tempfile)])
                                                else:
                                                        datasets_to_pass[-1].append(pickle.load(tempfile))
                                                tempfile.close()
                                else:
                                        if not tag['path']==open_file:
                                                open_ds = pydicom.dcmread(tag['path'])
                                                open_file = tag['path']
                                                #~ gc.collect()
                                        if not tag['enhanced']:
                                                if new_series:
                                                        datasets_to_pass.append([open_ds])
                                                else:
                                                        datasets_to_pass[-1].append(open_ds)
                                        else:
                                                split_ds = get_frame_ds(tag['instance'],open_ds)
                                                if new_series:
                                                        datasets_to_pass.append([split_ds])
                                                else:
                                                        datasets_to_pass[-1].append(split_ds)

                                                # Removed instruction to save temp ds in MIPPY 2.0 due to problems with pickling
                                                # the new dataset object.  Can only pickle if loaded from disk??
                                                #~ io.save_temp_ds(split_ds,tempdir,tag['instanceuid']+'.mds')
                                previous_tag = tag
        else:
                for tag in list_of_tags:
                        if tag['instanceuid'] in uids_to_match:
                                dcm_info.append((tag['instanceuid'],tag['path'],tag['instance']))
                gc.collect()
                f = partial(get_dataset,tempdir=tempdir)
                datasets_to_pass = multithread(f,dcm_info)
                # Group by series, to be flattened later if 1D list required
                datasets_to_pass = [list(g) for k,g, in itertools.groupby(datasets_to_pass, lambda ds: ds.SeriesInstanceUID)]

        return datasets_to_pass
